from pathlib import Path
from typing import List

from nipype import config

from clinica.pipelines.dwi.preprocessing.engine import DWIPreprocessingPipeline

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DwiPreprocessingUsingPhaseDiffFMap(DWIPreprocessingPipeline):
    """DWI Preprocessing using phase difference fieldmap.

    Ideas for improvement:
        - Use promising sdcflows workflows and/or dMRIprep

    Note:
        Some reading regarding the reproducibility of FSL eddy command:
        https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=fsl;1ccf038f.1608

    Returns:
        A clinica pipeline object containing the DwiPreprocessingUsingPhaseDiffFMap pipeline.
    """

    @staticmethod
    def get_processed_images(
        caps_directory: Path, subjects: List[str], sessions: List[str]
    ) -> List[str]:
        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import DWI_PREPROC_NII
        from clinica.utils.inputs import clinica_file_reader

        image_ids: List[str] = []
        if caps_directory.is_dir():
            preproc_files, _ = clinica_file_reader(
                subjects, sessions, caps_directory, DWI_PREPROC_NII, False
            )
            image_ids = extract_image_ids(preproc_files)
        return image_ids

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        List[str] :
            The list of inputs for the DwiPreprocessingUsingPhaseDiffFMap pipeline namely:
                - "dwi_filename" : The path of the diffusion weighted image in BIDS format.
                - "b_vectors_filename" : The path of the b-vectors file in BIDS format.
                - "b_values_filename": The path of the b-values file in BIDS format.
                - "dwi_json_filename" : The path of the DWI JSON file in BIDS format and containing
                  'TotalReadoutTime' and 'PhaseEncodingDirection' metadata (see BIDS specifications).
                - "fmap_magnitude_filename": The path to the (1st) magnitude image in BIDS format.
                - "fmap_phasediff_filename" : The path of the phase difference image in BIDS format.
                - "fmap_phasediff_json_filename" : The path of the phase difference JSON file in BIDS
                  format and containing 'EchoTime1' & 'EchoTime2' metadata (see BIDS specifications).
        """
        return [
            "dwi_filename",
            "b_vectors_filename",
            "b_values_filename",
            "dwi_json_filename",
            "fmap_magnitude_filename",
            "fmap_phasediff_filename",
            "fmap_phasediff_json_filename",
        ]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns:
            List[str]: The list of outputs generated by the DwiPreprocessingUsingPhaseDiffFMap pipeline namely:
                * preproc_dwi: Path of the preprocessed DWI
                * preproc_bvec: Path of the preprocessed bvec
                * preproc_bval: Path of the preprocessed bval
                * b0_mask: Path of the b0 brainmask
                * magnitude_on_b0: Path of the smoothed calibrated fmap on b0 space
                * calibrated_fmap_on_b0: Path of the calibrated fmap on b0 space
                * smoothed_fmap_on_b0: Path of the magnitude fmap on b0 space
        """
        return [
            "preproc_dwi",
            "preproc_bvec",
            "preproc_bval",
            "b0_mask",
            "magnitude_on_b0",
            "calibrated_fmap_on_b0",
            "smoothed_fmap_on_b0",
        ]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.input_files import (
            DWI_BVAL,
            DWI_BVEC,
            DWI_JSON,
            DWI_NII,
            FMAP_MAGNITUDE1_NII,
            FMAP_PHASEDIFF_JSON,
            FMAP_PHASEDIFF_NII,
        )
        from clinica.utils.inputs import clinica_list_of_files_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        list_bids_files = clinica_list_of_files_reader(
            self.subjects,
            self.sessions,
            self.bids_directory,
            [
                DWI_NII,
                DWI_BVEC,
                DWI_BVAL,
                DWI_JSON,
                FMAP_MAGNITUDE1_NII,
                FMAP_PHASEDIFF_NII,
                FMAP_PHASEDIFF_JSON,
            ],
            raise_exception=True,
        )
        save_participants_sessions(
            self.subjects, self.sessions, self.base_dir / self.name
        )
        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                f"List available in {self.base_dir / self.name / 'participants.tsv'}"
            )
            cprint(
                "Computational time will depend of the number of volumes in your DWI dataset and the use of CUDA."
            )

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                (name, value)
                for name, value in zip(self.get_input_fields(), list_bids_files)
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_node, self.input_node, [(x, x) for x in self.get_input_fields()]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.nipype import container_from_filename, fix_join

        from .tasks import rename_into_caps_task

        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="container_path",
        )

        rename_into_caps = npe.Node(
            nutil.Function(
                input_names=[
                    "dwi_filename",
                    "dwi_preproc_filename",
                    "b_values_preproc_filename",
                    "b_vectors_preproc_filename",
                    "b0_brain_mask_filename",
                    "calibrated_magnitude_image_filename",
                    "calibrated_field_map_image_filename",
                    "calibrated_smoothed_field_map_image_filename",
                ],
                output_names=[
                    "out_caps_dwi",
                    "out_caps_bval",
                    "out_caps_bvec",
                    "out_caps_brainmask",
                    "out_caps_magnitude",
                    "out_caps_fmap",
                    "out_caps_smoothed_fmap",
                ],
                function=rename_into_caps_task,
            ),
            name="rename_into_caps",
        )

        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = str(self.caps_directory)
        write_results.inputs.parameterization = False

        self.connect(
            [
                (
                    self.input_node,
                    container_path,
                    [("dwi_filename", "bids_or_caps_filename")],
                ),
                (self.input_node, rename_into_caps, [("dwi_filename", "dwi_filename")]),
                (
                    self.output_node,
                    rename_into_caps,
                    [
                        ("preproc_dwi", "dwi_preproc_filename"),
                        ("preproc_bval", "b_values_preproc_filename"),
                        ("preproc_bvec", "b_vectors_preproc_filename"),
                        ("b0_mask", "b0_brain_mask_filename"),
                        ("magnitude_on_b0", "calibrated_magnitude_image_filename"),
                        (
                            "calibrated_fmap_on_b0",
                            "calibrated_field_map_image_filename",
                        ),
                        (
                            "smoothed_fmap_on_b0",
                            "calibrated_smoothed_field_map_image_filename",
                        ),
                    ],
                ),
                (
                    container_path,
                    write_results,
                    [(("container", fix_join, "dwi"), "container")],
                ),
                (
                    rename_into_caps,
                    write_results,
                    [
                        ("out_caps_dwi", "preprocessing.@preproc_dwi"),
                        ("out_caps_bval", "preprocessing.@preproc_bval"),
                        ("out_caps_bvec", "preprocessing.@preproc_bvec"),
                        ("out_caps_brainmask", "preprocessing.@b0_mask"),
                        ("out_caps_magnitude", "preprocessing.@magnitude"),
                        ("out_caps_fmap", "preprocessing.@fmap"),
                        ("out_caps_smoothed_fmap", "preprocessing.@smoothed_fmap"),
                    ],
                ),
            ]
        )

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix3 as mrtrix3
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.pipelines.dwi.preprocessing.tasks import compute_average_b0_task
        from clinica.pipelines.dwi.preprocessing.workflows import eddy_fsl_pipeline

        from .utils import init_input_node, print_end_pipeline
        from .workflows import calibrate_and_register_fmap, compute_reference_b0

        init_node = npe.Node(
            interface=nutil.Function(
                input_names=self.get_input_fields(),
                output_names=[
                    "image_id",
                    "dwi_filename",
                    "b_vectors_filename",
                    "b_values_filename",
                    "total_readout_time",
                    "phase_encoding_direction",
                    "fmap_magnitude_filename",
                    "fmap_phasediff_filename",
                    "delta_echo_time",
                ],
                function=init_input_node,
            ),
            name="0-InitNode",
        )

        reference_b0 = compute_reference_b0(
            self.base_dir,
            b_value_threshold=self.parameters["low_bval"],
            use_cuda=self.parameters["use_cuda"],
            initrand=self.parameters["initrand"],
        )
        fmap_calibration_and_registration = calibrate_and_register_fmap(
            base_dir=self.base_dir
        )

        # Step 3: Run FSL eddy
        # ====================
        eddy = eddy_fsl_pipeline(
            self.base_dir,
            use_cuda=self.parameters["use_cuda"],
            initrand=self.parameters["initrand"],
            image_id=True,
            field=True,
            compute_mask=False,
        )

        # Step 4: Bias correction
        # =======================
        # Use implementation detailed in (Jeurissen et al., 2014)
        bias = npe.Node(mrtrix3.DWIBiasCorrect(use_ants=True), name="4-RemoveBias")

        # Step 5: Final brainmask
        # =======================
        # Compute average b0 on corrected dataset (for brain mask extraction)
        compute_avg_b0 = npe.Node(
            nutil.Function(
                input_names=["dwi_filename", "b_value_filename"],
                output_names=["reference_b0"],
                function=compute_average_b0_task,
            ),
            name="5a-ComputeB0Average",
        )
        compute_avg_b0.inputs.b_value_threshold = self.parameters["low_bval"]

        # Compute b0 mask on corrected avg b0
        mask_avg_b0 = npe.Node(fsl.BET(mask=True, robust=True), name="5b-MaskB0")

        # Print end message
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["image_id", "final_file"], function=print_end_pipeline
            ),
            name="99-WriteEndMessage",
        )

        connections = [
            (self.input_node, init_node, [(x, x) for x in self.get_input_fields()]),
            (
                init_node,
                reference_b0,
                [
                    (x, f"inputnode.{x}")
                    for x in (
                        "b_values_filename",
                        "b_vectors_filename",
                        "dwi_filename",
                        "total_readout_time",
                        "phase_encoding_direction",
                        "image_id",
                    )
                ],
            ),
            (
                init_node,
                fmap_calibration_and_registration,
                [
                    ("fmap_magnitude_filename", "inputnode.bias_magnitude_fmap"),
                    ("fmap_phasediff_filename", "inputnode.fmap_phasediff"),
                    ("delta_echo_time", "inputnode.delta_echo_time"),
                ],
            ),
            (
                reference_b0,
                fmap_calibration_and_registration,
                [("outputnode.reference_b0", "inputnode.reference_b0")],
            ),
            (
                init_node,
                eddy,
                [
                    (x, f"inputnode.{x}")
                    for x in (
                        "dwi_filename",
                        "b_values_filename",
                        "b_vectors_filename",
                        "image_id",
                        "total_readout_time",
                        "phase_encoding_direction",
                    )
                ],
            ),
            (
                fmap_calibration_and_registration,
                eddy,
                [("outputnode.smooth_calibrated_fmap", "inputnode.field")],
            ),
            (reference_b0, eddy, [("outputnode.brainmask", "inputnode.in_mask")]),
            # Step 4: Bias correction
            # =======================
            (init_node, bias, [("b_values_filename", "in_bval")]),
            (
                eddy,
                bias,
                [
                    ("outputnode.out_rotated_bvecs", "in_bvec"),
                    ("outputnode.out_corrected", "in_file"),
                ],
            ),
            # Step 5: Final brainmask
            # =======================
            # Compute average b0 on corrected dataset (for brain mask extraction)
            (init_node, compute_avg_b0, [("b_values_filename", "b_value_filename")]),
            (bias, compute_avg_b0, [("out_file", "dwi_filename")]),
            # Compute b0 mask on corrected avg b0
            (compute_avg_b0, mask_avg_b0, [("reference_b0", "in_file")]),
            # Print end message
            (init_node, print_end_message, [("image_id", "image_id")]),
            (mask_avg_b0, print_end_message, [("mask_file", "final_file")]),
            # Output node
            (init_node, self.output_node, [("b_values_filename", "preproc_bval")]),
            (
                eddy,
                self.output_node,
                [("outputnode.out_rotated_bvecs", "preproc_bvec")],
            ),
            (bias, self.output_node, [("out_file", "preproc_dwi")]),
            (mask_avg_b0, self.output_node, [("mask_file", "b0_mask")]),
            (
                fmap_calibration_and_registration,
                self.output_node,
                [
                    (
                        "outputnode.bet_magnitude_fmap_registered_onto_b0",
                        "magnitude_on_b0",
                    )
                ],
            ),
            (
                fmap_calibration_and_registration,
                self.output_node,
                [("outputnode.registered_calibrated_fmap", "calibrated_fmap_on_b0")],
            ),
            (
                fmap_calibration_and_registration,
                self.output_node,
                [("outputnode.smooth_calibrated_fmap", "smoothed_fmap_on_b0")],
            ),
        ]

        self.connect(connections)



from cmath import nan
from pathlib import Path
from typing import Iterable, List, Optional

import pandas as pd

__all__ = [
    "find_clinical_data",
    "read_imaging_data",
    "merge_imaging_and_clinical_data",
    "prepare_dataset_to_bids_format",
    "write_bids",
]


def find_clinical_data(clinical_data_directory: Path) -> pd.DataFrame:
    from clinica.utils.stream import cprint

    image_data_file = [f for f in clinical_data_directory.glob("*.csv")]
    if len(image_data_file) == 0:
        raise FileNotFoundError(
            f"Clinical data not found or incomplete in {clinical_data_directory}."
        )
    if len(image_data_file) > 1:
        raise ValueError(
            f"Too many data files found in {clinical_data_directory}, expected only one."
        )
    df_clinical = pd.read_csv(
        image_data_file[0],
        usecols=["eid", "31-0.0", "34-0.0", "21022-0.0", "21003-2.0", "21003-3.0"],
    )
    cprint(msg="All clinical data have been found", lvl="info")
    required_columns = {
        "eid",
        "31-0.0",
        "34-0.0",
        "21022-0.0",
        "21003-2.0",
        "21003-3.0",
    }
    if (
        len(missing_columns := required_columns.difference(set(df_clinical.columns)))
        > 0
    ):
        raise ValueError(f"Required columns {missing_columns} not found")
    df_clinical = df_clinical.rename(
        columns={
            "31-0.0": "sex_f31_0_0",
            "34-0.0": "year_of_birth_f34_0_0",
            "21022-0.0": "age_at_recruitment_f21022_0_0",
            "21003-2.0": "age_when_attended_assessment_centre_f21003_2_0",
            "21003-3.0": "age_when_attended_assessment_centre_f21003_3_0",
        }
    )
    return df_clinical


def read_imaging_data(imaging_data_directory: Path) -> pd.DataFrame:
    source_path_series_nifti = pd.Series(
        _find_imaging_data(imaging_data_directory), name="source_path"
    )
    source_path_series_dicom = pd.Series(
        _find_dicom_data(imaging_data_directory), name="source_path"
    )
    # Rationales for this choice of files can be found in the documentation:
    # https://github.com/aramis-lab/clinica/blob/dev/docs/Converters/UKBtoBIDS.md
    # list of the files we want to build the bids for each modality
    file_mod_list = [
        "T1.nii.gz",
        "T2_FLAIR_orig_defaced.nii.gz",
        "AP.nii.gz",
        "PA.nii.gz",
        "rfMRI.dcm",
        "tfMRI.dcm",
        "SWI.nii.gz",
    ]

    dataframe_nifti = pd.DataFrame.from_records(
        source_path_series_nifti, columns=["source_zipfile", "source_filename"]
    )
    dataframe_nifti = dataframe_nifti[
        dataframe_nifti["source_filename"].str.contains("unusable") != True  # noqa
    ]
    dataframe_dicom = pd.DataFrame.from_records(
        source_path_series_dicom, columns=["source_zipfile", "source_filename"]
    )
    dataframe = pd.concat([dataframe_nifti, dataframe_dicom])
    filename = dataframe["source_filename"].apply(lambda x: x.name).rename("filename")
    dataframe = dataframe[filename.isin(file_mod_list)]
    filename = filename[filename.isin(file_mod_list)]
    if dataframe.empty:
        raise ValueError(
            f"No imaging data were found in the provided folder: {imaging_data_directory}, "
            "or they are not handled by Clinica. Please check your data."
        )
    dataframe["source_zipfile"] = dataframe["source_zipfile"].apply(lambda x: str(x))
    split_zipfile = dataframe["source_zipfile"].str.split("_", expand=True)
    split_zipfile = split_zipfile.rename(
        {0: "source_id", 1: "modality_num", 2: "source_sessions_number"}, axis="columns"
    )  # .drop_duplicates()
    split_zipfile = split_zipfile.astype({"source_id": "int64"})

    df_source = pd.concat(
        [
            dataframe,
            filename,
            split_zipfile,
        ],
        axis=1,
    )
    return df_source


def _find_dicom_data(path_to_source_data: Path) -> Iterable[list[Path, Path]]:
    """
    This function finds the paths to the dicoms, only for fMRI, since we use the niftis for the other modalities.
    More information here: https://github.com/aramis-lab/clinica/blob/dev/docs/Converters/UKBtoBIDS.md
    """
    for z in path_to_source_data.rglob("*.zip"):
        if z.name.split("_")[1] == "20217":
            yield [z.relative_to(path_to_source_data), Path("fMRI/tfMRI.dcm")]
        elif z.name.split("_")[1] == "20225":
            yield [z.relative_to(path_to_source_data), Path("fMRI/rfMRI.dcm")]


def _find_imaging_data(path_to_source_data: Path) -> Iterable[list[Path, Path]]:
    from zipfile import ZipFile

    for z in path_to_source_data.rglob("*.zip"):
        for f in ZipFile(z).namelist():
            if f.endswith(".nii.gz"):
                yield [z.relative_to(path_to_source_data), Path(f)]


def merge_imaging_and_clinical_data(
    imaging_data: pd.DataFrame, clinical_data: pd.DataFrame
) -> pd.DataFrame:
    """Merges the imaging and clinical dataframes based on the subject id."""
    merged_data = clinical_data.merge(
        imaging_data, how="inner", right_on="source_id", left_on="eid"
    )
    return merged_data


def _complete_clinical(df_clinical: pd.DataFrame) -> pd.DataFrame:
    """This function uses the existing data to create the columns needed for
    the bids hierarchy (subject_id, ses, age_at _sessions, etc.)"""
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    df_clinical = df_clinical.assign(
        participant_id=lambda df: df.source_id.astype("str").apply(
            lambda x: bids_id_factory(StudyName.UKB).from_original_study_id(x)
        )
    )
    df_clinical = df_clinical.assign(
        sessions=lambda df: "ses-" + df.source_sessions_number.astype("str")
    )
    df_clinical = df_clinical.join(
        df_clinical.filename.map(
            {
                "T2_FLAIR_orig_defaced.nii.gz": {
                    "datatype": "anat",
                    "modality": "FLAIR",
                    "suffix": "FLAIR",
                    "sidecars": ["T2_FLAIR/T2_FLAIR.json"],
                    "task": "",
                    "dir": "",
                },
                "T1.nii.gz": {
                    "datatype": "anat",
                    "modality": "T1W",
                    "suffix": "T1w",
                    "sidecars": ["T1/T1.json"],
                    "task": "",
                    "dir": "",
                },
                "AP.nii.gz": {
                    "datatype": "dwi",
                    "modality": "dwi",
                    "suffix": "dwi",
                    "sidecars": [
                        "dMRI/raw/AP.json",
                        "dMRI/raw/AP.bval",
                        "dMRI/raw/AP.bvec",
                    ],
                    "task": "",
                    "dir": "_dir-AP",
                },
                "PA.nii.gz": {
                    "datatype": "dwi",
                    "modality": "dwi",
                    "suffix": "dwi",
                    "sidecars": [
                        "dMRI/raw/PA.json",
                        "dMRI/raw/PA.bval",
                        "dMRI/raw/PA.bvec",
                    ],
                    "task": "",
                    "dir": "_dir-PA",
                },
                "rfMRI.dcm": {
                    "datatype": "func",
                    "modality": "rsfmri",
                    "suffix": "bold",
                    "sidecars": [],
                    "task": "_task-rest",
                    "dir": "",
                },
                "tfMRI.dcm": {
                    "datatype": "func",
                    "modality": "tfmri",
                    "suffix": "bold",
                    "sidecars": [],
                    "task": "_task-facesshapesemotion",
                    "dir": "",
                },
                "SWI.nii.gz": {
                    "datatype": "swi",
                    "modality": "swi",
                    "suffix": "swi",
                    "sidecars": [],
                    "task": "",
                    "dir": "",
                },
            }
        ).apply(pd.Series)
    )
    df_clinical = df_clinical.assign(year_of_birth=lambda df: df.year_of_birth_f34_0_0)

    df_clinical = df_clinical.assign(age=lambda df: df.age_at_recruitment_f21022_0_0)
    df_clinical = df_clinical.assign(
        age_at_first_session=lambda df: df.age_when_attended_assessment_centre_f21003_2_0
    )
    df_clinical["age_at_session"] = df_clinical.apply(
        lambda df: _select_sessions(df), axis=1
    )
    df_clinical = df_clinical[df_clinical["age_at_session"].notna()]
    df_clinical = df_clinical.assign(
        sessions_month=lambda df: (df.age_at_session - df.age_at_first_session) * 12
    )
    df_clinical = df_clinical.assign(
        sessions=lambda df: df.sessions_month.map(lambda x: f"ses-M{int(x):03d}")
    )
    df_clinical = df_clinical.assign(
        bids_filename=lambda df: (
            "sub-UKB"
            + df.source_id.astype("str")
            + "_"
            + df.sessions.astype("str")
            + df.dir.astype("str")
            + df.task.astype("str")
            + "_"
            + df.suffix
        )
    )
    df_clinical = df_clinical.join(
        df_clinical.sex_f31_0_0.astype("str")
        .map(
            {
                "0": {"sex": "F"},
                "1": {"sex": "M"},
            }
        )
        .apply(pd.Series)
    )
    df_clinical = df_clinical.assign(
        bids_full_path=lambda df: df.participant_id
        + "/"
        + df.sessions
        + "/"
        + df.datatype
        + "/"
        + df.bids_filename
    )
    return df_clinical


def prepare_dataset_to_bids_format(
    df_clinical: pd.DataFrame,
) -> dict[str, pd.DataFrame]:
    df_clinical = _complete_clinical(df_clinical)
    df_specifications = pd.read_csv(
        Path(__file__).parent / "specifications.csv", sep=";"
    )
    df_clinical = df_clinical.set_index(
        ["participant_id", "sessions", "modality", "bids_filename"],
        verify_integrity=True,
    )

    return {
        col: df_clinical.filter(items=list(df_specifications[col]))
        for col in ["participants", "sessions", "scans"]
    }


def write_bids(
    to: Path,
    participants: pd.DataFrame,
    sessions: pd.DataFrame,
    scans: pd.DataFrame,
    dataset_directory: Path,
) -> None:
    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
    from clinica.iotools.bids_utils import StudyName, write_to_tsv

    fs = LocalFileSystem(auto_mkdir=True)

    participants = participants.droplevel(
        ["sessions", "modality", "bids_filename"]
    ).drop_duplicates()

    # Ensure BIDS hierarchy is written first.
    with fs.transaction:
        with fs.open(
            str(to / "dataset_description.json"), "w"
        ) as dataset_description_file:
            BIDSDatasetDescription(name=StudyName.UKB).write(
                to=dataset_description_file
            )
        with fs.open(str(to / "participants.tsv"), "w") as participant_file:
            write_to_tsv(participants, participant_file)

    for participant_id, data_frame in sessions.groupby("participant_id"):
        sessions = data_frame.droplevel(
            ["participant_id", "modality", "bids_filename"]
        ).drop_duplicates()

        sessions_filepath = to / str(participant_id) / f"{participant_id}_sessions.tsv"
        with fs.open(str(sessions_filepath), "w") as sessions_file:
            write_to_tsv(sessions, sessions_file)
    scans = scans.set_index(["bids_full_path"], verify_integrity=True)
    for bids_full_path, metadata in scans.iterrows():
        if metadata["modality_num"] != "20217" and metadata["modality_num"] != "20225":
            _copy_file_to_bids(
                zipfile=dataset_directory / metadata["source_zipfile"],
                filenames=[metadata["source_filename"]] + metadata["sidecars"],
                bids_path=to / bids_full_path,
            )
        else:
            _convert_dicom_to_nifti(
                zipfiles=dataset_directory / metadata["source_zipfile"],
                bids_path=to / bids_full_path,
            )
            if metadata["modality_num"] == "20217":
                _import_event_tsv(bids_path=to)
    return


def _copy_file_to_bids(zipfile: Path, filenames: List[Path], bids_path: Path) -> None:
    """Install the requested files in the BIDS  dataset."""
    import fsspec

    fo = fsspec.open(str(zipfile))
    fs = fsspec.filesystem("zip", fo=fo)
    for filename in filenames:
        filename = str(filename)
        if fs.exists(filename):
            bids_path_extension = str(bids_path) + "." + (filename.split(".", 1)[1])
            with fsspec.open(bids_path_extension, mode="wb") as f:
                f.write(fs.cat(filename))


def _convert_dicom_to_nifti(zipfiles: Path, bids_path: Path) -> None:
    """Install the requested files in the BIDS  dataset.
    First, the dicom is extracted in a temporary directory
    Second, the dicom extracted is converted in the right place using dcm2niix"""
    import json
    import subprocess
    import tempfile
    import zipfile
    from pathlib import PurePath

    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)

    zf = zipfile.ZipFile(zipfiles)
    try:
        bids_path.parent.mkdir(exist_ok=True, parents=True)
    except OSError:
        # Folder already created with previous instance
        pass
    with tempfile.TemporaryDirectory() as tempdir:
        zf.extractall(tempdir)
        command = [
            "dcm2niix",
            "-w",
            "0",
        ]
        command += ["-9", "-z", "y"]
        command += ["-b", "y", "-ba", "y"]
        command += [tempdir]
        subprocess.run(command)
        fmri_image_path = _find_largest_imaging_data(Path(tempdir))
        fmri_image_path = fmri_image_path or ""
        fs.copy(str(fmri_image_path), str(bids_path) + ".nii.gz")
        fs.copy(
            str(fmri_image_path.parent)
            + "/"
            + PurePath(PurePath(fmri_image_path.name).stem).stem
            + ".json",
            str(bids_path) + ".json",
        )
    # Add the taskname to the json
    with open(str(bids_path) + ".json", "r+") as f:
        json_file = json.load(f)
        json_file["TaskName"] = "facesshapesemotion"
        f.seek(0)
        json.dump(json_file, f, indent=4)


def _select_sessions(x: pd.DataFrame) -> Optional[pd.Series]:
    from clinica.utils.stream import cprint

    if (
        x["source_sessions_number"] == "2"
        and x.age_when_attended_assessment_centre_f21003_2_0 != nan
    ):
        return x.age_when_attended_assessment_centre_f21003_2_0
    elif (
        x["source_sessions_number"] == "2"
        and x.age_when_attended_assessment_centre_f21003_2_0 == nan
    ):
        cprint(
            msg=(
                f"The subject {x.eid} doesn't have the age for the imaging session "
                "number one (age_when_attended_assessment_centre_f21003_2_0)."
                f"It will not be converted. To have it converted, please update your clinical data."
            ),
            lvl="warning",
        )
        return None
    elif (
        x["source_sessions_number"] == "3"
        and x.age_when_attended_assessment_centre_f21003_3_0 != nan
    ):
        return x.age_when_attended_assessment_centre_f21003_3_0
    elif (
        x["source_sessions_number"] == "3"
        and x.age_when_attended_assessment_centre_f21003_3_0 == nan
    ):
        cprint(
            msg=(
                f"The subject {x.eid} doesn't have the age for the imaging session "
                "number two (age_when_attended_assessment_centre_f21003_3_0)."
                f"It will not be converted. To have it converted, please update your clinical data."
            ),
            lvl="warning",
        )
        return None


def _import_event_tsv(bids_path: Path) -> None:
    """Import the csv containing the events' information."""
    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)
    event_tsv = (
        Path(__file__).parents[3]
        / "resources"
        / "fmri"
        / "task-facesshapesemotion_events.tsv"
    )
    if not event_tsv.exists():
        raise FileNotFoundError(
            f"Could not find file {event_tsv} which is an internal file of Clinica."
        )
    bids_path_extension = bids_path / "task-facesshapesemotion_events.tsv"
    fs.copy(str(event_tsv), str(bids_path_extension))


def _find_largest_imaging_data(path_to_source_data: Path) -> Optional[Path]:
    import os

    weight = 0
    file_to_return = None
    for z in path_to_source_data.rglob("*.nii.gz"):
        if os.path.getsize(z) > weight:
            weight = os.path.getsize(z)
            file_to_return = z
    if file_to_return:
        return file_to_return
    return None



"""Methods used by BIDS converters."""

import json
import os
import re
from abc import ABC, abstractmethod
from collections import UserString
from enum import Enum
from pathlib import Path
from typing import BinaryIO, List, Optional, Type, Union

import pandas as pd

from clinica.utils.pet import Tracer


class StudyName(str, Enum):
    """Studies supported by the converters of Clinica."""

    ADNI = "ADNI"
    AIBL = "AIBL"
    GENFI = "GENFI"
    HABS = "HABS"
    NIFD = "NIFD"
    OASIS = "OASIS"
    OASIS3 = "OASIS3"
    UKB = "UKB"
    IXI = "IXI"


BIDS_VALIDATOR_CONFIG = {
    "ignore": [
        # Possibly dcm2nii(x) errors
        "NIFTI_UNIT",
        "INCONSISTENT_PARAMETERS",
        # fMRI-specific errors
        "SLICE_TIMING_NOT_DEFINED",
        "NIFTI_PIXDIM4",
        "BOLD_NOT_4D",
        "REPETITION_TIME_MUST_DEFINE",
        "TASK_NAME_MUST_DEFINE",
        # Won't fix errors
        "MISSING_SESSION",  # Allows subjects to have different sessions
        "INCONSISTENT_SUBJECTS",  # Allows subjects to have different modalities
        "SCANS_FILENAME_NOT_MATCH_DATASET",  # Necessary until PET is added to BIDS standard
        "CUSTOM_COLUMN_WITHOUT_DESCRIPTION",  # We won't create these JSON files as clinical description
        # is already done in TSV files of clinica.
        "NO_AUTHORS",  # Optional field in dataset_description.json
    ],
    "warn": [],
    "error": [],
    "ignoredFiles": [],
}


class BIDSSubjectID(ABC, UserString):
    """This is the interface that BIDS subject IDs have to implement."""

    def __init__(self, value: str):
        instance = super().__init__(self.validate(value))
        return instance

    @abstractmethod
    def validate(self, value: str) -> str:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_original_study_id(cls, study_id: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def to_original_study_id(self) -> str:
        raise NotImplementedError


def bids_id_factory(study: StudyName) -> Type[BIDSSubjectID]:
    if study == StudyName.ADNI:
        return ADNIBIDSSubjectID
    if study == StudyName.NIFD:
        return NIFDBIDSSubjectID
    if study == StudyName.AIBL:
        return AIBLBIDSSubjectID
    if study == StudyName.UKB:
        return UKBBIDSSubjectID
    if study == StudyName.GENFI:
        return GENFIBIDSSubjectID
    if study == StudyName.OASIS:
        return OASISBIDSSubjectID
    if study == StudyName.OASIS3:
        return OASIS3BIDSSubjectID
    if study == StudyName.HABS:
        return HABSBIDSSubjectID
    if study == StudyName.IXI:
        return IXIBIDSSubjectID


class ADNIBIDSSubjectID(BIDSSubjectID):
    """Implementation for ADNI of the BIDSSubjectIDClass, allowing to go from the source id XXX_S_XXXX
    to a bids id sub-ADNIXXXSXXX and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-ADNI\d{3}S\d{4}", value):
            return value
        raise ValueError(
            f"BIDS ADNI subject ID {value} is not properly formatted. "
            "Expecting a 'sub-ADNIXXXSXXXX' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"\d{3}_S_\d{4}", study_id):
            return "sub-ADNI" + study_id.replace("_", "")
        raise ValueError(
            f"Raw ADNI subject ID {study_id} is not properly formatted. "
            "Expecting a 'XXX_S_XXXX' format."
        )

    def to_original_study_id(self) -> str:
        return "_S_".join(self.split("ADNI")[1].split("S"))


class NIFDBIDSSubjectID(BIDSSubjectID):
    """Implementation for NIFD of the BIDSSubjectIDClass, allowing to go from the source id X_S_XXXX
    to a bids id sub-NIFDXSXXX and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-NIFD\dS\d{4}", value):
            return value
        raise ValueError(
            f"BIDS NIFD subject ID {value} is not properly formatted. "
            "Expecting a 'sub-NIFDXSXXXX' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"\d_S_\d{4}", study_id):
            return "sub-NIFD" + study_id.replace("_", "")
        raise ValueError(
            f"Raw NIFD subject ID {study_id} is not properly formatted. "
            "Expecting a 'X_S_XXXX' format."
        )

    def to_original_study_id(self) -> str:
        return "_S_".join(self.split("NIFD")[1].split("S"))


class AIBLBIDSSubjectID(BIDSSubjectID):
    """Implementation for AIBL of the BIDSSubjectIDClass, allowing to go from the source id Y
    to a bids id sub-ADNIY and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-AIBL\d*", value):
            return value
        raise ValueError(
            f"BIDS AIBL subject ID {value} is not properly formatted. "
            "Expecting a 'sub-AIBLY' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"\d*", study_id):
            return "sub-AIBL" + study_id
        raise ValueError(
            f"Raw AIBL subject ID {study_id} is not properly formatted. "
            "Expecting a 'Y' format where Y is a combination of digits."
        )

    def to_original_study_id(self) -> str:
        return self.split("AIBL")[1]


class UKBBIDSSubjectID(BIDSSubjectID):
    """Implementation for UKB of the BIDSSubjectIDClass, allowing to go from the source id Y
    to a bids id sub-ADNIY and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-UKB\d*", value):
            return value
        raise ValueError(
            f"BIDS UKB subject ID {value} is not properly formatted. "
            "Expecting a 'sub-UKBY' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"\d*", study_id):
            return "sub-UKB" + study_id
        raise ValueError(
            f"Raw UKB subject ID {study_id} is not properly formatted. "
            "Expecting a 'Y' format where Y is a combination of digits."
        )

    def to_original_study_id(self) -> str:
        return self.split("UKB")[1]


class GENFIBIDSSubjectID(BIDSSubjectID):
    """Implementation for GENFI of the BIDSSubjectIDClass, allowing to go from the source id Y
    to a bids id sub-Y and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-\w*", value):
            return value
        raise ValueError(
            f"BIDS GENFI subject ID {value} is not properly formatted. "
            "Expecting a 'sub-Y' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"\w*", study_id):
            return "sub-" + study_id
        raise ValueError(
            f"Raw GENFI subject ID {study_id} is not properly formatted. "
            "Expecting a 'Y' format where Y is a combination of letters and digits."
        )

    def to_original_study_id(self) -> str:
        return self.split("-")[1]


class OASISBIDSSubjectID(BIDSSubjectID):
    """Implementation for OASIS1 of the BIDSSubjectIDClass, allowing to go from the source id OAS1_XXXX_MR1/2
    to a bids id sub-OASIS1XXXX and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-OASIS1\d{4}", value):
            return value
        raise ValueError(
            f"BIDS OASIS1 subject ID {value} is not properly formatted. "
            "Expecting a 'sub-OASIS1XXXX' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"OAS1_\d{4}_MR\d", study_id):
            return "sub-OASIS1" + study_id.split("_")[1]
        raise ValueError(
            f"Raw OASIS1 subject ID {study_id} is not properly formatted. "
            "Expecting a 'OAS1_XXXX_MR1/2' format."
        )

    def to_original_study_id(self) -> str:
        return f"OAS1_{self.split('OASIS1')[1]}_MR1"


class OASIS3BIDSSubjectID(BIDSSubjectID):
    """Implementation for OASIS3 of the BIDSSubjectIDClass, allowing to go from the source id XXXX
    to a bids id sub-OAS3XXXX and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-OAS3\d{4}", value):
            return value
        raise ValueError(
            f"BIDS OASIS3 subject ID {value} is not properly formatted. "
            "Expecting a 'sub-OAS3XXXX' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"OAS3\d{4}", study_id):
            return "sub-" + study_id
        raise ValueError(
            f"Raw OASIS3 subject ID {study_id} is not properly formatted. "
            "Expecting a 'OAS3XXXX' format."
        )

    def to_original_study_id(self) -> str:
        return self.split("-")[1]


class HABSBIDSSubjectID(BIDSSubjectID):
    """Implementation for HABS of the BIDSSubjectIDClass, allowing to go from the source id P_Y
    to a bids id sub-HABSY and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-HABS\w*", value):
            return value
        raise ValueError(
            f"BIDS HABS subject ID {value} is not properly formatted. "
            "Expecting a 'sub-HABSY' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"P_\w*", study_id):
            return study_id.replace("P_", "sub-HABS")
        raise ValueError(
            f"Raw HABS subject ID {study_id} is not properly formatted. "
            "Expecting a 'P_Y' format."
        )

    def to_original_study_id(self) -> str:
        return str(self).replace("sub-HABS", "P_")


class IXIBIDSSubjectID(BIDSSubjectID):
    """Implementation for IXI of the BIDSSubjectIDClass, allowing to go from the source id IXI###
    to a bids id sub-IXI### and reciprocally."""

    def validate(self, value: str) -> str:
        if re.fullmatch(r"sub-IXI\d{3}", value):
            return value
        raise ValueError(
            f"BIDS IXI subject ID {value} is not properly formatted. "
            "Expecting a 'sub-IXIXXX' format."
        )

    @classmethod
    def from_original_study_id(cls, study_id: str) -> str:
        if re.fullmatch(r"IXI\d{3}", study_id):
            return f"sub-{study_id}"
        raise ValueError(
            f"Raw IXI subject ID {study_id} is not properly formatted. "
            "Expecting a 'Y' format."
        )

    def to_original_study_id(self) -> str:
        return str(self).replace("sub-", "")


# -- Methods for the clinical data --
def create_participants_df(
    study_name: StudyName,
    clinical_specifications_folder: Path,
    clinical_data_dir: Path,
    bids_ids: list[str],
    delete_non_bids_info: bool = True,
) -> pd.DataFrame:
    """Create the file participants.tsv.

    Parameters
    ----------
    study_name : StudyName
        The name of the study (Ex. ADNI).

    clinical_specifications_folder : Path
        The path to the clinical file.

    clinical_data_dir : Path
        The path to the directory where the clinical data are stored.

    bids_ids : list of str
        The list of bids ids.

    delete_non_bids_info : bool, optional
        If True delete all the rows of the subjects that are not available in the BIDS dataset.
        Default=True.

    Returns
    -------
    pd.DataFrame :
        A pandas dataframe that contains the participants data.
    """
    import numpy as np

    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv
    from clinica.utils.stream import cprint

    fields_bids = ["participant_id"]
    prev_location = ""
    prev_sheet = 0
    index_to_drop = []
    subjects_to_drop = []
    study_name = StudyName(study_name)
    location_name = f"{study_name.value} location"

    participants_specs = pd.read_csv(
        clinical_specifications_folder / "participant.tsv", sep="\t"
    )
    participant_fields_db = participants_specs[study_name.value]
    field_location = participants_specs[location_name]
    participant_fields_bids = participants_specs["BIDS CLINICA"]

    # Extract the list of the available BIDS fields for the dataset
    for i in range(0, len(participant_fields_db)):
        if not pd.isnull(participant_fields_db[i]):
            fields_bids.append(participant_fields_bids[i])

    # Init the dataframe that will be saved in the file participants.tsv
    participant_df = pd.DataFrame(columns=fields_bids)

    for i in range(0, len(participant_fields_db)):
        # If a field not empty is found
        if not pd.isnull(participant_fields_db[i]):
            # Extract the file location of the field and read the value from the file
            tmp = field_location[i].split("/")
            location = tmp[0]
            # If a sheet is available
            sheet = tmp[1] if len(tmp) > 1 else 0
            # Check if the file to open for a certain field is the same of the previous field
            if location == prev_location and sheet == prev_sheet:
                pass
            else:
                file_ext = os.path.splitext(location)[1]
                file_to_read_path = clinical_data_dir / location
                if file_ext == ".xlsx":
                    file_to_read = pd.read_excel(file_to_read_path, sheet_name=sheet)
                elif file_ext == ".csv":
                    file_to_read = load_clinical_csv(
                        clinical_data_dir, location.split(".")[0]
                    )
                    # Condition to handle ADNI modification of file APOERES.csv
                    # See issue https://github.com/aramis-lab/clinica/issues/1294
                    if study_name == StudyName.ADNI and location == "APOERES.csv":
                        if (
                            participant_fields_db[i] not in file_to_read.columns
                            and "GENOTYPE" in file_to_read.columns
                        ):
                            # Split the 'GENOTYPE' column into 'APGEN1' and 'APGEN2'
                            genotype = file_to_read["GENOTYPE"].str.split(
                                "/", expand=True
                            )
                            file_to_read = file_to_read.assign(
                                APGEN1=genotype[0], APGEN2=genotype[1]
                            )

                prev_location = location
                prev_sheet = sheet

            field_col_values = []
            # For each field in fields_dataset extract all the column values
            for j in range(0, len(file_to_read)):
                # Convert the alternative_id_1 to string if is an integer/float
                value_to_read = file_to_read[participant_fields_db[i]]
                if participant_fields_bids[i] == "alternative_id_1" and (
                    value_to_read.dtype == np.float64 or value_to_read.dtype == np.int64
                ):
                    if not pd.isnull(file_to_read.at[j, participant_fields_db[i]]):
                        value_to_append = str(
                            file_to_read.at[j, participant_fields_db[i]]
                        ).rstrip(".0")
                    else:
                        value_to_append = np.NaN
                else:
                    value_to_append = file_to_read.at[j, participant_fields_db[i]]
                field_col_values.append(value_to_append)
            # Add the extracted column to the participant_df
            participant_df[participant_fields_bids[i]] = pd.Series(field_col_values)

    if study_name == StudyName.ADNI or study_name == StudyName.AIBL:
        # ADNImerge contains one row for each visits so there are duplicates
        participant_df = participant_df.drop_duplicates(
            subset=["alternative_id_1"], keep="first"
        )
    elif study_name == StudyName.OASIS:
        # OASIS provides several MRI for the same session
        participant_df = participant_df[
            ~participant_df.alternative_id_1.str.endswith("_MR2")
        ]
    participant_df.reset_index(inplace=True, drop=True)

    # Adding participant_id column with BIDS ids
    for i in range(0, len(participant_df)):
        value = bids_id_factory(study_name).from_original_study_id(
            participant_df["alternative_id_1"][i]
        )
        bids_id = [s for s in bids_ids if value in s]
        if len(bids_id) == 0:
            index_to_drop.append(i)
            subjects_to_drop.append(value)
        else:
            participant_df.at[i, "participant_id"] = bids_id[0]

    if len(subjects_to_drop) > 0:
        cprint(
            msg=(
                "The following subjects of dataset directory were not found in your BIDS folder :\n"
                + ", ".join(subjects_to_drop)
            ),
            lvl="info",
        )
    # Delete all the rows of the subjects that are not available in the BIDS dataset
    if delete_non_bids_info:
        participant_df = participant_df.drop(index_to_drop)

    participant_df = participant_df.fillna("n/a")

    return participant_df


def create_scans_dict(
    clinical_data_dir: Path,
    study_name: StudyName,
    clinical_specifications_folder: Path,
    bids_ids: list[str],
    name_column_ids: str,
    name_column_ses: str,
    ses_dict: dict,
) -> pd.DataFrame:
    """[summary].

    Parameters
    ----------
    clinical_data_dir : Path
        The path to the directory where the clinical data are stored.

    study_name : StudyName
        The name of the study (Ex ADNI).

    clinical_specifications_folder : Path
        The path to the folder containing the clinical specification files.

    bids_ids : list of str
        A list of bids ids.

    name_column_ids : str
        The name of the column where the subject id is contained.

    name_column_ses : str
        The name of the column where the viscode of the session is contained.

    ses_dict : dict
        The links the session id to the viscode of the session.

    Returns
    -------
    pd.DataFrame :
        A pandas DataFrame that contains the scans information for all sessions of all participants.
    """
    import datetime

    from clinica.utils.pet import Tracer
    from clinica.utils.stream import cprint

    scans_dict = {}
    prev_file = ""
    prev_sheet = ""

    # Init the dictionary with the subject ids
    for bids_id in bids_ids:
        scans_dict[bids_id] = dict()
        for session_id in {"ses-" + key for key in ses_dict[bids_id].keys()}:
            scans_dict[bids_id][session_id] = {
                "T1/DWI/fMRI/FMAP": {},
                Tracer.PIB: {},
                Tracer.AV45: {},
                Tracer.FMM: {},
                Tracer.FDG: {},
            }

    scans_specs = pd.read_csv(clinical_specifications_folder / "scans.tsv", sep="\t")
    fields_dataset = []
    fields_location = []
    fields_bids = []
    fields_mod = []

    # Extract the fields available and the corresponding bids name, location and type
    for i in range(0, len(scans_specs[study_name.value])):
        field = scans_specs[study_name.value][i]
        if not pd.isnull(field):
            fields_dataset.append(field)
            fields_bids.append(scans_specs["BIDS CLINICA"][i])
            fields_location.append(scans_specs[f"{study_name.value} location"][i])
            fields_mod.append(scans_specs["Modalities related"][i])

    # For each field available extract the original name, extract from the file all the values and fill a data structure
    for i in range(0, len(fields_dataset)):
        # Location is composed by file/sheet
        location = fields_location[i].split("/")
        file_name = location[0]
        sheet = location[1] if len(location) > 1 else ""
        # Check if the file to read is already opened
        if file_name == prev_file and sheet == prev_sheet:
            pass
        else:
            file_ext = os.path.splitext(file_name)[1]
            files_to_read = [f for f in clinical_data_dir.glob(file_name)]
            if file_ext == ".xlsx":
                file_to_read = pd.read_excel(files_to_read[0], sheet_name=sheet)
            elif file_ext == ".csv":
                file_path = files_to_read[0]

                # Fix for malformed flutemeta file in AIBL (see #796).
                # Some flutemeta lines contain a non-coded string value at the second-to-last position. This value
                # contains a comma which adds an extra column and shifts the remaining values to the right. In this
                # case, we just remove the erroneous content and replace it with -4 which AIBL uses as n/a value.
                on_bad_lines = (  # noqa: E731
                    lambda bad_line: bad_line[:-3] + [-4, bad_line[-1]]
                    if "flutemeta" in file_path and study_name == StudyName.AIBL
                    else "error"
                )

                file_to_read = pd.read_csv(
                    file_path,
                    sep=",",
                    engine="python",
                    on_bad_lines=on_bad_lines,
                )
            prev_file = file_name
            prev_sheet = sheet

        for bids_id in bids_ids:
            original_id = bids_id.replace(f"sub-{study_name.value}", "")
            for session_name in {"ses-" + key for key in ses_dict[bids_id].keys()}:
                # When comparing sessions, remove the "-ses" prefix IF it exists
                row_to_extract = file_to_read[
                    (file_to_read[name_column_ids] == int(original_id))
                    & (
                        list(
                            filter(
                                None, file_to_read[name_column_ses].str.split("ses-")
                            )
                        )[0][0]
                        == ses_dict[bids_id][
                            list(filter(None, session_name.split("ses-")))[0]
                        ]
                    )
                ].index.tolist()
                if len(row_to_extract) > 0:
                    row_to_extract = row_to_extract[0]
                    # Fill the dictionary with all the information
                    value = file_to_read.iloc[row_to_extract][fields_dataset[i]]

                    if study_name == StudyName.AIBL:  # Deal with special format in AIBL
                        if value == "-4":
                            value = "n/a"
                        elif fields_bids[i] == "acq_time":
                            date_obj = datetime.datetime.strptime(value, "%m/%d/%Y")
                            value = date_obj.strftime("%Y-%m-%dT%H:%M:%S")

                    scans_dict[bids_id][session_name][fields_mod[i]][
                        fields_bids[i]
                    ] = value
                else:
                    cprint(
                        f"Scans information for {bids_id} {session_name} not found.",
                        lvl="info",
                    )
                    scans_dict[bids_id][session_name][fields_mod[i]][
                        fields_bids[i]
                    ] = "n/a"

    return scans_dict


def _write_bids_dataset_description(
    study_name: StudyName,
    bids_dir: Path,
    bids_version: Optional[str] = None,
) -> None:
    """Write `dataset_description.json` at the root of the BIDS directory."""
    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription

    if bids_version:
        bids_desc = BIDSDatasetDescription(name=study_name, bids_version=bids_version)
    else:
        bids_desc = BIDSDatasetDescription(name=study_name)
    with open(bids_dir / "dataset_description.json", "w") as f:
        bids_desc.write(to=f)


def _write_readme(
    study_name: StudyName,
    data_dict: dict,
    bids_dir: Path,
) -> None:
    """Write `README` at the root of the BIDS directory."""
    from clinica.iotools.bids_readme import BIDSReadme

    bids_readme = BIDSReadme(
        name=study_name,
        link=data_dict["link"],
        description=data_dict["desc"],
    )
    with open(bids_dir / "README", "w") as f:
        bids_readme.write(to=f)


def _write_bids_validator_config(bids_dir: Path) -> None:
    """Write `.bids-validator-config.json` at the root of the BIDS directory."""
    with open(bids_dir / ".bids-validator-config.json", "w") as f:
        json.dump(BIDS_VALIDATOR_CONFIG, f, skipkeys=True, indent=4)


def _write_bidsignore(bids_dir: Path) -> None:
    """Write `.bidsignore` file at the root of the BIDS directory."""
    with open(bids_dir / ".bidsignore", "w") as f:
        # pet/ is necessary until PET is added to BIDS standard
        f.write("\n".join(["swi/\n"]))
        f.write("\n".join(["conversion_info/"]))


def write_modality_agnostic_files(
    study_name: StudyName,
    readme_data: dict,
    bids_dir: Path,
    bids_version: Optional[str] = None,
) -> None:
    """
    Write the files README, dataset_description.json, .bidsignore and .bids-validator-config.json
    at the root of the BIDS directory.

    Parameters
    ----------
    study_name : StudyName
        The name of the study (Ex ADNI).

    readme_data : dict
        A dictionary containing the data specific to the dataset to write in the readme.

    bids_dir : Path
        The path to the bids directory.

    bids_version : str, optional
        The BIDS version if different from the version supported by Clinica.
    """
    _write_bids_dataset_description(study_name, bids_dir, bids_version)
    _write_readme(study_name, readme_data, bids_dir)
    _write_bids_validator_config(bids_dir)
    _write_bidsignore(bids_dir)


def _get_pet_tracer_from_filename(filename: str) -> Tracer:
    """Return the PET tracer from the provided filename.

    Parameters
    ----------
    filename : str
        The filename from which to extract the PET tracer.

    Returns
    -------
    tracer : Tracer
        The PET tracer.

    Raises
    ------
    ValueError
        If no tracer found in the filename.
    """
    import re

    tracer = None
    for entity in ("trc", "acq"):
        m = re.search(rf"({entity}-[a-zA-Z0-9]+)", filename)
        if m:
            tracer = m.group(1)[4:].upper()
    if tracer is None:
        raise ValueError(
            f"Could not extract the PET tracer from the following file name {filename}."
        )

    return Tracer(tracer)


def write_scans_tsv(
    bids_dir: Path, participant_ids: List[str], scans_dict: dict
) -> None:
    """Write the scans dict into TSV files.

    Parameters
    ----------
    bids_dir : Path
        The path to the BIDS directory.

    participant_ids : List[str]
        List of participant ids for which to write the scans TSV files.

    scans_dict : dict
        Dictionary containing scans metadata.

        .. note::
            This is the output of the function
            `clinica.iotools.bids_utils.create_scans_dict`.

    See also
    --------
    write_sessions_tsv
    """
    supported_modalities = ("anat", "dwi", "func", "pet")

    for sub in participant_ids:
        for session_path in (bids_dir / sub).glob("ses-*"):
            scans_df = pd.DataFrame()
            tsv_file = (
                bids_dir
                / sub
                / session_path.name
                / f"{sub}_{session_path.name}_scans.tsv"
            )
            tsv_file.unlink(missing_ok=True)

            for mod in (bids_dir / sub / session_path.name).glob("*"):
                if mod.name in supported_modalities:
                    for file in [
                        file for file in mod.iterdir() if mod.suffix != ".json"
                    ]:
                        f_type = (
                            "T1/DWI/fMRI/FMAP"
                            if mod.name in ("anat", "dwi", "func")
                            else _get_pet_tracer_from_filename(file.name).value
                        )
                        row_to_append = pd.DataFrame(
                            scans_dict[sub][session_path.name][f_type], index=[0]
                        )
                        row_to_append.insert(
                            0, "filename", str(Path(mod.name) / Path(file.name))
                        )
                        scans_df = pd.concat([scans_df, row_to_append])
            scans_df = scans_df.set_index("filename").fillna("n/a")
            scans_df.to_csv(tsv_file, sep="\t", encoding="utf8")


def get_bids_subjs_list(bids_path: Path) -> List[str]:
    """Given a BIDS compliant dataset, return the list of all the subjects available.

    Parameters
    ----------
    bids_path : Path
        The path to the BIDS directory.

    Returns
    -------
    List[str] :
        List of subject IDs available in this BIDS dataset.

    See also
    --------
    get_bids_sess_list
    get_bids_subjs_paths
    """
    return _filter_folder_names(bids_path, "sub-*")


def _filter_folder_names(folder: Path, pattern: str) -> List[str]:
    return [d.name for d in folder.glob(pattern) if d.is_dir()]


def get_bids_sess_list(subj_path: Path) -> List[str]:
    """Given a path to a subject's folder, this function returns the
    list of sessions available.

    Parameters
    ----------
    subj_path : Path
        The path to the subject folder for which to list the sessions.

    Returns
    -------
    List[str] :
        The list of session names for this subject.

    See also
    --------
    get_bids_subjs_list
    get_bids_subjs_paths
    """
    return _filter_folder_names(subj_path, "ses-*")


def get_bids_subjs_paths(bids_path: Path) -> List[Path]:
    """Given a BIDS compliant dataset, returns the list of all paths to the subjects folders.

    Parameters
    ----------
    bids_path : str
        Path to the BIDS directory.

    Returns
    -------
    List[Path] :
        List of paths to the subjects folders.

    See also
    --------
    get_bids_subjs_list
    get_bids_sess_list
    """
    return [d for d in bids_path.glob("sub-*") if d.is_dir()]


def remove_space_and_symbols(data: Union[str, List[str]]) -> Union[str, List[str]]:
    """Remove spaces, "-", and "_" characters from a string.

    If a list of strings is provided, this function will be called
    on each item of the list.

    Parameters
    ----------
    data : str or List[str]
        String or list of strings to be cleaned.

    Returns
    -------
    str or List[str] :
        Cleaned string(s).
    """
    import re

    if isinstance(data, list):
        return [remove_space_and_symbols(d) for d in data]
    return re.sub("[-_ ]", "", data)


def json_from_dcm(dcm_dir: Path, json_path: Path) -> None:
    """Writes descriptive JSON file from DICOM header.

    Parameters
    ----------
    dcm_dir : Path
        The path to the DICOM directory.

    json_path : Path
        The path to the output JSON file.
    """
    from pydicom import dcmread
    from pydicom.tag import Tag

    from clinica.utils.stream import cprint

    fields_dict = {
        "DeviceSerialNumber": Tag(("0018", "1000")),
        "Manufacturer": Tag(("0008", "0070")),
        "ManufacturersModelName": Tag(("0008", "1090")),
        "SoftwareVersions": Tag(("0018", "1020")),
        "BodyPart": Tag(("0018", "0015")),
        "Units": Tag(("0054", "1001")),
        # Institution
        "InstitutionName": Tag(("0008", "0080")),
        "InstitutionAddress": Tag(("0008", "0081")),
        "InstitutionalDepartmentName": Tag(("0008", "1040")),
        # MRI
        "MagneticFieldStrength": Tag(("0018", "0087")),
        # PET
        "InjectedRadioactivity": Tag(("0018", "1074")),
        "MolarActivity": Tag(("0018", "1077")),
        "InjectionStart": Tag(("0018", "1042")),
        "FrameDuration": Tag(("0018", "1242")),
    }

    try:
        dcm_path = [f for f in dcm_dir.glob("*.dcm")][0]
        ds = dcmread(dcm_path)
        json_dict = {
            key: ds.get(tag).value for key, tag in fields_dict.items() if tag in ds
        }
        with open(json_path, "w") as f:
            f.write(json.dumps(json_dict, skipkeys=True, indent=4))
    except IndexError:
        cprint(msg=f"No DICOM found at {dcm_dir}", lvl="warning")


def _build_dcm2niix_command(
    input_dir: Path,
    output_dir: Path,
    output_fmt: str,
    compress: bool = False,
    bids_sidecar: bool = True,
) -> list:
    """Generate the dcm2niix command from user inputs."""
    command = ["dcm2niix", "-w", "0", "-f", output_fmt, "-o", str(output_dir)]
    command += ["-9", "-z", "y"] if compress else ["-z", "n"]
    command += ["-b", "y", "-ba", "y"] if bids_sidecar else ["-b", "n"]
    command += [str(input_dir)]

    return command


def run_dcm2niix(
    input_dir: Path,
    output_dir: Path,
    output_fmt: str,
    compress: bool = False,
    bids_sidecar: bool = True,
) -> bool:
    """Runs the `dcm2niix` command using a subprocess.

    Parameters
    ----------
    input_dir : Path
        Input folder.

    output_dir : Path
        Output folder. This will be passed to the
        dcm2niix "-o" option.

    output_fmt : str
        Output format. This will be passed to the
        dcm2niix "-f" option.

    compress : bool, optional
        Whether to compress or not.
        Default=False.

    bids_sidecar : bool, optional
        Whether to generate a BIDS sidecar or not. Default=True.

    Returns
    -------
    bool :
        True if the conversion was successful, False otherwise.
    """
    import subprocess

    from clinica.utils.stream import cprint

    cprint(f"Attempting to convert {output_fmt}.", lvl="debug")
    command = _build_dcm2niix_command(
        input_dir, output_dir, output_fmt, compress, bids_sidecar
    )
    completed_process = subprocess.run(command, capture_output=True)

    if completed_process.returncode != 0:
        output_message = (
            completed_process.stdout.decode("utf-8")
            if completed_process.stdout is not None
            else ""
        )
        cprint(
            msg=(
                "The DICOM to BIDS conversion with dcm2niix failed:\n"
                f"command: {' '.join(command)}\n"
                f"{output_message}"
            ),
            lvl="warning",
        )
        return False
    cprint(
        msg=(
            f"The DICOM to BIDS conversion with dcm2niix was successful:\n"
            f"command: {' '.join(command)}\n"
        ),
        lvl="debug",
    )
    return True


# todo : place in genfi utils ?
def identify_modality(filename: str) -> Optional[str]:
    """Identifies the modality of a file given its name.

    Parameters
    ----------
    filename: str
        Input filename

    Returns
    -------
    Optional[str]:
        Modality or None if parsing uns
    """
    import numpy as np

    filename = filename.lower()
    if "dwi" in filename:
        return "dwi"
    if "t1" in filename:
        return "T1"
    if "t2" in filename:
        return "T2w"
    if "fieldmap" in filename:
        return "fieldmap"
    if "fmri" in filename:
        return "rsfmri"
    else:
        return np.nan


# todo : use more ?
def write_to_tsv(df: pd.DataFrame, buffer: Union[Path, BinaryIO]) -> None:
    """Save dataframe as a BIDS-compliant TSV file.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame to write to TSV file.

    buffer : PathLike or BinaryIO
        Either a file or a stream to write the DataFrame to.
    """
    df.to_csv(buffer, sep="\t", na_rep="n/a", date_format="%Y-%m-%d")



from typing import Optional

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "statistics-volume-correction"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.caps_directory
@click.argument("t_map", type=str)
@click.argument("height_threshold", type=float)
@click.argument("FWEp", type=float)
@click.argument("FDRp", type=float)
@click.argument("FWEc", type=float)
@click.argument("FDRc", type=float)
@cli_param.option_group.pipeline_specific_options
@cli_param.option_group.option(
    "-nc",
    "--n_cuts",
    default=8,
    show_default=True,
    help="Number of cuts along each direction",
)
@cli_param.option_group.common_pipelines_options
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
def cli(
    caps_directory: str,
    t_map: str,
    height_threshold: str,
    fwep: float,
    fdrp: float,
    fwec: float,
    fdrc: float,
    n_cuts: int = 8,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
) -> None:
    """Statistical correction of statistics-volume pipeline.

       T_MAP is the name of the T statistic map used for the correction.

       HEIGHT_THRESHOLD is the T value corresponding to an uncorrected p-value of 0.001.

       FWEp is the Family-Wise Error peak threshold.

       FDRp is the False Discovery Rate peak threshold.

       Finally, FWEc and FDRc stand for the respective correction parameters.

    Prerequisite: You have to perform the statistics-volume pipeline before running this pipeline.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/Stats_Volume/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .statistics_volume_correction_pipeline import StatisticsVolumeCorrection

    parameters = {
        "t_map": t_map,
        "height_threshold": height_threshold,
        "FWEp": fwep,
        "FDRp": fdrp,
        "FWEc": fwec,
        "FDRc": fdrc,
        "n_cuts": n_cuts,
    }

    pipeline = StatisticsVolumeCorrection(
        caps_directory=caps_directory,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



"""This module contains utilities for longitudinal pipelines. See CAPS specifications for details about long ID."""
from pathlib import Path
from typing import List, Optional


def get_long_id(session_ids: List[str]) -> str:
    """Extract longitudinal ID from a list of session IDs.

    This will create a unique identifier for a participant and
    its corresponding sessions. Sessions labels are sorted alphabetically
    before being merged in order to generate the longitudinal ID.

    Parameters
    ----------
    session_ids : list of str
        List of session IDs (e.g. ["ses-M000"] or ["ses-M000", "ses-M018", "ses-M036"]).

    Returns
    -------
    str :
        Longitudinal ID.

    Examples
    --------
    >>> from clinica.utils.longitudinal import get_long_id
    >>> get_long_id(['ses-M000'])
    'long-M000'
    >>> get_long_id(['ses-M000', 'ses-M018', 'ses-M036'])
    'long-M000M018M036'
    >>> get_long_id(['ses-M018', 'ses-M036', 'ses-M000'])
    'long-M000M018M036'
    """
    if not all([session_id.startswith("ses-") for session_id in session_ids]):
        raise ValueError(
            "Expected a list of session IDs of the form ses-XXX, "
            f"but received {session_ids} instead."
        )
    return "long-" + "".join(
        [session_id.lstrip("ses-") for session_id in sorted(session_ids)]
    )


def get_participants_long_id(
    participant_ids: List[str], session_ids: List[str]
) -> List[str]:
    """Extract list of longitudinal IDs from a set of participant and session IDs.

    Parameters
    ----------
    participant_ids : list of str
        List of participant IDs for which to compute the longitudinal IDs.

    session_ids : list of str
        List of session IDs for which to compute the longitudinal IDs.

    Returns
    -------
    list of str :
        The computed longitudinal IDs.

    Examples
    --------
    >>> from clinica.utils.longitudinal import get_participants_long_id
    >>> get_participants_long_id(['sub-CLNC01', 'sub-CLNC01', 'sub-CLNC02'], ['ses-M000', 'ses-M018', 'ses-M000'])
    ['long-M000M018', 'long-M000M018', 'long-M000']
    """
    from .participant import get_unique_subjects

    _, sessions_for_each_subject = get_unique_subjects(participant_ids, session_ids)

    long_ids = []
    for sessions in sessions_for_each_subject:
        long_ids += [get_long_id(sessions)] * len(sessions)

    return long_ids


def save_long_id(
    session_ids: List[str],
    output_dir: Path,
    file_name: Optional[str] = None,
) -> None:
    """Save the list of session IDs to given `file_name`.

    Parameters
    ----------
    session_ids : list of str
        The list of session IDs to save.

    output_dir : Path
        The path to the output directory in which to save the session IDs.

    file_name : str, optional
        The file name to use. If None, this will be computed as:
        '{longitudinal_id}_sessions.tsv'
    """
    if not output_dir.exists():
        output_dir.mkdir(parents=True)
    file_name = file_name or f"{get_long_id(session_ids)}_sessions.tsv"
    content = "\n".join(sorted(session_ids))
    with open(output_dir / file_name, "w") as fp:
        fp.write(f"session_id\n{content}\n")


def read_sessions(
    caps_dir: Path,
    participant_id: str,
    longitudinal_id: str,
) -> List[str]:
    """Extract sessions IDs from `caps_dir`/subjects/`participant_id`/`long_id`/`long_id`_sessions.tsv.

    Parameters
    ----------
    caps_dir : Path
        Path to CAPS folder.

    participant_id : str
        ID of subject for which to extract session IDs.

    longitudinal_id : str
        Longitudinal ID for which to extract session IDs.

    Returns
    -------
    List of str :
        The extracted list of session IDs.

    Raises
    ------
    ClinicaException :
        If expected session TSV file does not exist.
        If 'session_id' is not in the session dataframe.
    """
    import pandas as pd

    from clinica.utils.exceptions import ClinicaException

    sessions_file = (
        caps_dir
        / "subjects"
        / participant_id
        / longitudinal_id
        / f"{longitudinal_id}_sessions.tsv"
    )

    if not sessions_file.is_file():
        raise ClinicaException(
            "The TSV file with sessions associated "
            f"to {participant_id} for longitudinal ID {longitudinal_id} is missing "
            f"(expected path: {sessions_file})."
        )
    df = pd.read_csv(sessions_file, sep="\t")
    if "session_id" not in df.columns:
        raise ClinicaException(
            f"The TSV file does not contain session_id column (path: {sessions_file})."
        )

    return list(df.session_id)



__all__ = [
    "init_input_node_task",
    "save_to_caps_task",
    "write_tsv_files_task",
]


def init_input_node_task(t1w: str, recon_all_args: str, output_dir: str) -> tuple:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.t1.utils import init_input_node

    image_id, t1w, flags, subjects_dir = init_input_node(
        Path(t1w), recon_all_args, Path(output_dir)
    )

    return image_id, str(t1w), flags, str(subjects_dir)


def save_to_caps_task(
    source_dir: str, image_id: str, caps_dir: str, overwrite_caps: bool = False
) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.t1.utils import save_to_caps

    return save_to_caps(
        Path(source_dir), image_id, Path(caps_dir), overwrite_caps=overwrite_caps
    )


def write_tsv_files_task(subjects_dir: str, image_id: str) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.t1.utils import write_tsv_files

    return write_tsv_files(Path(subjects_dir), image_id)



from pathlib import Path
from typing import Optional, Tuple, Union

from clinica.utils.pet import SUVRReferenceRegion

_all__ = [
    "init_input_node",
    "concatenate_transforms",
    "perform_suvr_normalization",
    "rename_into_caps",
    "print_end_pipeline",
]


def init_input_node(pet: str) -> str:
    """Initiate the pipeline."""
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    image_id = get_subject_id(pet)
    print_begin_image(image_id)
    return pet


def concatenate_transforms(
    pet_to_t1w_transform: str, t1w_to_mni_transform: str
) -> list:
    """Concatenate two input transformation files into a list.

    Parameters
    ----------
    pet_to_t1w_transform : str
        First transformation to apply.

    t1w_to_mni_transform : str
        Second transformation to apply.

    Returns
    -------
    list :
        Both transform files path in a list.
    """
    return [t1w_to_mni_transform, pet_to_t1w_transform]


def perform_suvr_normalization(
    pet_image_path: Path,
    normalizing_image_path: Path,
    reference_mask_path: Path,
) -> Path:
    """Normalize the input image according to the reference region.

    It uses nilearn `resample_to_img` and scipy `trim_mean` functions.
    This function is different than the one in other PET pipelines
    because there is a down-sampling step.

    Parameters
    ----------
    pet_image_path : Path
        The path to the image to be processed.

    normalizing_image_path : Path
        The path to the image used to compute the mean of the reference region.

    reference_mask_path : Path
        The path to the mask of the reference region.

    Returns
    -------
    output_img : Path
        The path to the normalized nifti image.
    """
    import nibabel as nib
    import numpy as np
    from nilearn.image import resample_to_img
    from scipy.stats import trim_mean

    from clinica.utils.filemanip import get_filename_no_ext
    from clinica.utils.stream import cprint

    pet_image = nib.load(pet_image_path)
    normalizing_image = nib.load(normalizing_image_path)
    reference_mask = nib.load(reference_mask_path)

    # Down-sample the pet image used for normalization so we can multiply it with the mask
    down_sampled_image = resample_to_img(
        normalizing_image, reference_mask, interpolation="nearest"
    )

    # Compute the mean of the region
    region = np.multiply(
        down_sampled_image.get_fdata(), reference_mask.get_fdata(dtype="float32")
    )
    array_region = np.where(region != 0, region, np.nan).flatten()
    region_mean = trim_mean(array_region[~np.isnan(array_region)], 0.1)
    cprint(region_mean, lvl="info")

    # Divide the value of the image voxels by the computed mean
    data = pet_image.get_fdata(dtype="float32") / region_mean

    # Create and save the normalized image
    output_image = (
        Path.cwd() / f"{get_filename_no_ext(pet_image_path)}_suvr_normalized.nii.gz"
    )
    normalized_img = nib.Nifti1Image(data, pet_image.affine, header=pet_image.header)
    normalized_img.to_filename(output_image)

    return output_image


def rename_into_caps(
    pet_bids_image_filename: Path,
    pet_preprocessed_image_filename: Path,
    pet_to_mri_transformation_filename: Path,
    suvr_reference_region: Union[str, SUVRReferenceRegion],
    uncropped_image: bool,
    pet_filename_in_t1w_raw: Optional[Path] = None,
    output_dir: Optional[Path] = None,
) -> Tuple[Path, Path, Optional[Path]]:
    """
    Rename the outputs of the pipelines into CAPS format.

    More precisely, the following files will be renamed:

        - 'pet_filename_raw' will be renamed to 'pet_filename_caps'
        - 'transformation_filename_raw' will be renamed to 'transformation_filename_caps'
        - 'pet_filename_in_t1w_raw' will be renamed (if provided) to 'pet_filename_in_t1w_caps'

    Parameters
    ----------
    pet_bids_image_filename : Path
        Input PET image file from the BIDS dataset.
        This file is used to extract the entities from the source
        file like subject, session, run, tracer...

    pet_preprocessed_image_filename : Path
        Preprocessed PET file as outputted by Nipype.

    pet_to_mri_transformation_filename : Path
        Transformation file from PET to MRI space as outputted by Nipype.

    suvr_reference_region : str or SUVRReferenceRegion
        SUVR mask name for file name output.
        This will be used to derive the prefix of the PET image.

    uncropped_image : bool
        Pipeline argument for image cropping.
        This will be used to derive the prefix of the PET image.

    pet_filename_in_t1w_raw : str, optional
        Intermediate PET in T1w MRI space.
        If not provided, no renaming will be done.

    output_dir : str, optional
        Specify the output folder where the renamed files should
        be written. This is mostly used for testing purposes in order
        to enable this function to write to Pytest's temporary folders.
        When used in the pipeline, this is left as None as other nodes
        are responsible for adding the missing pieces to the output paths
        and writing the files to disk.

    Returns
    -------
    pet_filename_caps : Path
        The renamed preprocessed PET file to match CAPS conventions.

    transformation_filename_caps : Path
        The transformation file from PET to MRI space renamed to match CAPS conventions.

    pet_filename_in_t1w_caps : Path or None
        Intermediate PET in T1w MRI space renamed to match CAPS conventions.
        If 'pet_filename_in_t1w_raw' is None, this will be None.
    """
    bids_entities = _get_bids_entities_without_suffix(
        pet_bids_image_filename, suffix="pet"
    )
    if output_dir:
        bids_entities = str(output_dir / bids_entities)
    pet_filename_caps = _rename_pet_into_caps(
        bids_entities,
        pet_preprocessed_image_filename,
        not uncropped_image,
        SUVRReferenceRegion(suvr_reference_region),
    )
    transformation_filename_caps = _rename_transformation_into_caps(
        bids_entities, pet_to_mri_transformation_filename
    )
    pet_filename_in_t1w_caps = None
    if pet_filename_in_t1w_raw is not None:
        pet_filename_in_t1w_caps = _rename_intermediate_pet_in_t1w_space_into_caps(
            bids_entities, pet_filename_in_t1w_raw
        )

    return pet_filename_caps, transformation_filename_caps, pet_filename_in_t1w_caps


def _get_bids_entities_without_suffix(filename: Path, suffix: str) -> str:
    """Return the BIDS entities without the suffix from a BIDS path."""
    from clinica.utils.filemanip import get_filename_no_ext

    return get_filename_no_ext(filename).rstrip(f"_{suffix}")


def _rename_pet_into_caps(
    entities: str, filename: Path, cropped: bool, region: SUVRReferenceRegion
) -> Path:
    """Rename into CAPS PET."""
    return _rename(filename, entities, _get_pet_bids_components(cropped, region))


def _rename_transformation_into_caps(entities: str, filename: Path) -> Path:
    """Rename into CAPS transformation file."""
    return _rename(filename, entities, "_space-T1w_rigid.mat")


def _rename_intermediate_pet_in_t1w_space_into_caps(
    entities: str, filename: Path
) -> Path:
    """Rename intermediate PET in T1w MRI space."""
    return _rename(filename, entities, "_space-T1w_pet.nii.gz")


def _rename(filename: Path, entities: str, suffix: str) -> Path:
    """Rename 'filename' into '{entities}{suffix}'."""
    from nipype.interfaces.utility import Rename

    rename = Rename()
    rename.inputs.in_file = str(filename)
    rename.inputs.format_string = f"{entities}{suffix}"

    return Path(rename.run().outputs.out_file)


def _get_pet_bids_components(cropped: bool, region: SUVRReferenceRegion) -> str:
    """Return a string composed of the PET-specific entities (space, resolution,
    desc, and suvr), suffix, and extension.
    """
    space = "_space-MNI152NLin2009cSym"
    resolution = "_res-1x1x1"
    desc = "_desc-Crop" if cropped else ""
    suvr = f"_suvr-{region.value}"

    return f"{space}{desc}{resolution}{suvr}_pet.nii.gz"


def print_end_pipeline(pet: str, final_file):
    """
    Display end message for <subject_id> when <final_file> is connected.
    """
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(pet))



from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple

import pandas as pd
import pydicom as pdcm
from pandas import DataFrame

__all__ = [
    "merge_imaging_and_clinical_data",
    "parse_clinical_data",
    "parse_imaging_data",
    "prepare_dataset_to_bids_format",
    "write_bids",
]


def _find_dicoms(path_to_source_data: Path) -> Iterable[Tuple[Path, Path]]:
    """Find the dicoms in the given directory.

    Parameters
    ----------
    path_to_source_data: Path
        The path to the source data.

    Returns
    -------
    Iterable[Tuple[Path, Path]]
        Path to found files and parent directory
    """
    from clinica.utils.stream import cprint

    cprint("Looking for imaging data.", lvl="info")
    for z in path_to_source_data.rglob("*.dcm"):
        yield z, z.parent


def _filter_dicoms(df: DataFrame) -> DataFrame:
    """Filters modalities handled by the converter.

    Parameters
    ----------
    df: DataFrame
        Dataframe containing all of the dicoms available in the source directory.

    Returns
    -------
    df: DataFrame
        Dataframe with only the modalities handled.
    """
    to_filter = [
        # from GENFI 1
        "t2_2d_axial",
        "dwi_trace",
        "dwi_adc",
        "dwi_fa",
        "localizer",
        "localiser",
        "nonimage",
        "dual_pd_t2_axial",
        "t1_1.25mm_iso",
        "MoCo",
        # from GENFI 2
        "ASL",
        "motioncorrected",
        "localiser",
        "localizer",
    ]

    df = df.drop_duplicates(subset=["source"])
    df = df.assign(
        series_desc=lambda x: x.source_path.apply(
            lambda y: _handle_series_description(y)
        ),
        acq_date=lambda x: x.source_path.apply(lambda y: pdcm.dcmread(y).StudyDate),
        manufacturer=lambda x: x.source_path.apply(lambda y: _handle_manufacturer(y)),
    )
    df = df.set_index(["source_path"], verify_integrity=True)
    df = df[df["source"].str.contains("secondary", case=False) != True]  # noqa
    for file_mod in to_filter:
        df = df[~df["series_desc"].str.contains(file_mod, case=False)]
    return df


def _handle_series_description(x: str) -> str:
    """This function handles the series description used to identify the modality later on.

    In the case of T1 and T2, some dicoms are misnamed. The echo time and repetition time
    are supposed to be very different between a T1 and T2. Therefore, we use this proxy to check that the series description is coherent.
    The values used are in milliseconds, and have been chosen arbitrarily.
    """
    import warnings

    MAX_ECHO_TIME_FOR_A_T1 = 60
    MAX_REPETITION_TIME_FOR_A_T1 = 2100
    series_description = pdcm.dcmread(x).SeriesDescription
    if any([modality in series_description.lower() for modality in ["t1", "t2"]]):
        try:
            echo_time = pdcm.dcmread(x).EchoTime
            repetition_time = pdcm.dcmread(x).RepetitionTime
            if (
                "t2" in series_description.lower()
                and echo_time < MAX_ECHO_TIME_FOR_A_T1
                and repetition_time < MAX_REPETITION_TIME_FOR_A_T1
            ):
                return "t1"
        except AttributeError:
            warnings.warn(
                message=(
                    f"The subject from DICOM {Path(x).parent} has no Echo Time "
                    "or Repetition Time, it might be wrongly converted."
                )
            )
    return series_description


def _handle_manufacturer(x: str) -> str:
    try:
        return pdcm.dcmread(x).Manufacturer
    except Exception:
        return "Unknown"


def _check_file(directory: Path, pattern: str) -> Path:
    try:
        data_file = [f for f in directory.glob(pattern)]
    except StopIteration:
        raise FileNotFoundError("Clinical data file not found.")
    if len(data_file) == 0:
        raise FileNotFoundError("Clinical data not found or incomplete. Aborting")
    if len(data_file) > 1:
        raise ValueError("Too many data files found, expected one. Aborting.")
    return data_file[0]


def parse_clinical_data(clinical_data_directory: Path) -> DataFrame:
    return _complete_clinical_data(*_find_clinical_data(clinical_data_directory))


def _find_clinical_data(
    clinical_data_directory: Path,
) -> tuple[DataFrame, DataFrame, DataFrame, DataFrame, DataFrame]:
    """Finds the clinical data associated with the dataset.

    Parameters
    ----------
    clinical_data_directory: Path
        The path to the clinical data.

    Returns
    -------
    List[DataFrame]
        Dataframes containing the clinical data
    """
    from clinica.utils.stream import cprint

    cprint("Looking for clinical data.", lvl="info")

    return tuple(
        _read_file(_check_file(clinical_data_directory, pattern))
        for pattern in (
            "FINAL*DEMOGRAPHICS*.xlsx",
            "FINAL*IMAGING*.xlsx",
            "FINAL*CLINICAL*.xlsx",
            "FINAL*BIOSAMPLES*.xlsx",
            "FINAL*NEUROPSYCH*.xlsx",
        )
    )


def _read_file(data_file: Path) -> pd.DataFrame:
    return (
        pd.concat(
            [
                pd.read_excel(data_file),
                pd.read_excel(data_file, sheet_name=1),
            ]
        )
        .convert_dtypes()
        .rename(columns=lambda x: x.lower().replace(" ", "_"))
    )


def _complete_clinical_data(
    df_demographics: DataFrame,
    df_imaging: DataFrame,
    df_clinical: DataFrame,
    df_biosamples: DataFrame,
    df_neuropsych: DataFrame,
) -> DataFrame:
    """Merges the different clincal dataframes into one.

    Parameters
    ----------
    df_demographics: DataFrame
        Dataframe containing the demographic data

    df_imaging: DataFrame
        Dataframe containing the imaging data

    df_clinical: DataFrame
        Dataframe containing the clinical data

    df_biosamples: DataFrame
        Dataframe containing the biosample data

    df_neuropsych: DataFrame
        Dataframe containing the neuropsych data

    Returns
    -------
    df_clinical_complete: DataFrame
        Dataframe with the data of the 3 input dataframes
    """
    merge_key = ["blinded_code", "blinded_site", "visit"]
    df_clinical_complete = df_imaging.merge(
        df_demographics, how="inner", on=merge_key
    ).drop(columns="diagnosis")
    df_clinical_complete = df_clinical_complete.merge(
        df_biosamples, how="inner", on=merge_key
    )
    df_clinical_complete = df_clinical_complete.merge(
        df_neuropsych, how="inner", on=merge_key
    )
    df_clinical = df_clinical.dropna(subset=merge_key)
    return df_clinical_complete.merge(df_clinical, how="inner", on=merge_key)


def prepare_dataset_to_bids_format(
    complete_data_df: DataFrame,
    gif: bool,
    path_to_clinical_tsv: Path,
) -> Dict[str, DataFrame]:
    """Selects the data needed to write the participants, sessions, and scans tsvs.

    Parameters
    ----------
    complete_data_df: DataFrame
        Dataframe containing the merged data extracted from the raw images and the clinical data

    gif: bool
        If True, indicates the user wants to have the values of the gif parcellation

    path_to_clinical_tsv: Path
        TSV file containing the data fields the user wishes to have from the excel spreadsheets

    Returns
    -------
    Dict[str, DataFrame]
        Dictionary containing as key participants, sessions and scans, and the values wanted for each tsv
    """
    complete_data_df = complete_data_df.drop_duplicates(
        subset=["participant_id", "session_id", "modality", "run_num", "bids_filename"]
    ).set_index(
        ["participant_id", "session_id", "modality", "run_num", "bids_filename"],
        verify_integrity=True,
    )
    specifications = pd.read_csv(Path(__file__).parent / "specifications.csv", sep=";")
    if not gif:
        specifications = specifications.head(8)
    # add additional data through csv
    if path_to_clinical_tsv:
        additional_data_df = pd.read_csv(path_to_clinical_tsv, sep="\t")
        data_mapping = pd.read_csv(Path(__file__) / "data_mapping.tsv", sep="\t")
        pre_addi_df = data_mapping.merge(additional_data_df, how="inner", on="data")
        addi_df = pd.DataFrame(
            [
                pre_addi_df["data"][pre_addi_df["dest"] == x].values.tolist()
                for x in ("participants", "sessions", "scans")
            ]
        ).transpose()
        addi_df.columns = ["participants", "sessions", "scans"]
        df_to_write = pd.concat([specifications, addi_df])
    else:
        df_to_write = specifications
    return {
        col: complete_data_df.filter(items=list(df_to_write[col]))
        for col in ["participants", "sessions", "scans"]
    }


def merge_imaging_and_clinical_data(
    imaging_data: DataFrame, clinical_data: DataFrame
) -> DataFrame:
    """This function merges the dataframe containing the data extracted
    from the raw images and from the clinical data.

    Parameters
    ----------
    imaging_data : DataFrame
        Dataframe containing the data extracted from the raw images

    clinical_data : DataFrame
        Dataframe containing the clinical data

    Returns
    -------
    df_complete : DataFrame
        Dataframe containing the merged data
    """
    df_complete = imaging_data.merge(
        clinical_data,
        how="inner",
        left_on=["source_id", "source_ses_id"],
        right_on=["blinded_code", "visit"],
    )
    return df_complete.loc[:, ~df_complete.columns.duplicated()]


def parse_imaging_data(source_path: Path) -> DataFrame:
    return _merge_imaging_data(_read_imaging_data(source_path))


def _read_imaging_data(source_path: Path) -> DataFrame:
    """This function finds the imaging data and filters it.

    Parameters
    ----------
    source_path: Path
        The path to the raw data.

    Returns
    -------
    df_dicom: DataFrame
        Dataframe containing the data extracted.
    """
    return _filter_dicoms(
        pd.DataFrame(_find_dicoms(source_path), columns=["source_path", "source"])
    )


def _merge_imaging_data(df: DataFrame) -> DataFrame:
    """This function uses the raw information extracted from the images,
    to obtain all the information necessary for the BIDS conversion.

    Parameters
    ----------
    df: DataFrame
        Dataframe containing the data extracted from the images

    Returns
    -------
    DataFrame :
        Dataframe with the data necessary for the BIDS
    """
    df = _compute_source_id_and_source_ses_id(df).reset_index()
    df = _compute_genfi_version(df)
    df = _compute_baseline_and_session_numbers(df)
    df = _compute_participant_id(df)
    df = _compute_modality(df)
    df = _compute_fieldmaps(df)
    df = _compute_runs(df)
    df = _compute_bids_full_path(df)

    return df


def _compute_source_id_and_source_ses_id(df: DataFrame) -> DataFrame:
    """Adds two columns built from the column 'source'.

    - 'source': subject ID and session ID joined by a dash
    - 'source_id': subject ID in the raw dataset
    - 'source_ses_id': Session ID in the raw dataset

    Example
    -------
    source = "'C9ORF001-11'
    source_id = 'C9ORF001'
    source_ses_id = '11
    """
    from clinica.utils.filemanip import get_parent

    return df.assign(
        source_id=lambda x: x.source.apply(
            lambda y: get_parent(y, 2).name.split("-")[0]
        ),
        source_ses_id=lambda x: x.source.apply(
            lambda y: get_parent(y, 2).name.split("-")[1]
        ),
    )


def _compute_genfi_version(df: DataFrame) -> DataFrame:
    """Compute the genfi_version from the souce_ses_id column.

    The column `source_ses_id` is casted to integer values.
    """
    return df.assign(
        source_ses_id=lambda x: x.source_ses_id.astype("int"),
        genfi_version=lambda x: x.source_ses_id.apply(lambda y: f"GENFI{len(str(y))}"),
    )


def _compute_baseline_and_session_numbers(df: DataFrame) -> DataFrame:
    """Compute the baseline date and session numbers
    and merge on 'source_id` and `source_ses_id`.
    """
    df_with_acq_date = _compute_baseline_date(df)
    df_with_acq_date_and_session_numbers = _compute_session_numbers(df_with_acq_date)
    return df.merge(
        df_with_acq_date_and_session_numbers[["baseline", "session_id"]],
        how="inner",
        on=["source_id", "source_ses_id"],
    )


def _compute_baseline_date(df: DataFrame) -> DataFrame:
    """Computes the baseline date by taking the minimum
    acq_date for each subject.

    Parameters
    ----------
    df: Dataframe

    Returns
    -------
    Dataframe
        Contains the baseline date.
    """
    df_1 = (
        df[["source_id", "source_ses_id", "acq_date"]]
        .groupby(["source_id", "source_ses_id"])
        .min()
    )
    df_2 = df[["source_id", "acq_date"]].groupby("source_id").min()
    return df_1.join(df_2.rename(columns={"acq_date": "baseline"}))


def _compute_session_numbers(df: DataFrame) -> DataFrame:
    """Computes the session IDs obtained from the number of months between
    an acquisition and the baseline acquisition.

    Parameters
    ----------
    df: DataFrame
        Dataframe containing the timestamp of the acquisition.

    Returns
    -------
    DataFrame
        Dataframe containing the session_id computed from the timestamps.
    """
    from datetime import datetime

    for col in ("acq_date", "baseline"):
        df[col] = df[col].apply(lambda x: datetime.strptime(x, "%Y%m%d"))
    return df.assign(
        ses_month=lambda x: 12
        * (x.acq_date.apply(lambda y: y.year) - x.baseline.apply(lambda y: y.year))
        + (x.acq_date.apply(lambda y: y.month) - x.baseline.apply(lambda y: y.month)),
        session_id=lambda x: x.ses_month.map(lambda y: f"ses-M{y:03d}"),
    )


def _compute_participant_id(df: DataFrame) -> DataFrame:
    """Compute the 'participant_id' column from the 'source_id' column."""
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    return df.assign(
        participant_id=df.source_id.apply(
            lambda x: bids_id_factory(StudyName.GENFI).from_original_study_id(x)
        )
    )


def _compute_modality(df: DataFrame) -> DataFrame:
    """Parses the series_desc column to identify the modality, and map
    this modality to a metadata dictionary.

    Parameters
    ----------
    df: Dataframe
        DataFrame on which to compute the modality.

    Returns
    -------
    Dataframe
        Contains the datatype, suffix, sidecars and task that may be
        needed to form the complete file name.
    """
    from clinica.iotools.bids_utils import identify_modality

    modality_mapping = {
        "T2w": {
            "datatype": "anat",
            "suffix": "T2w",
            "sidecars": ["T2w.json"],
            "task": "",
        },
        "T1": {
            "datatype": "anat",
            "suffix": "T1w",
            "sidecars": ["T1w.json"],
            "task": "",
        },
        "dwi": {
            "datatype": "dwi",
            "suffix": "dwi",
            "sidecars": [
                "dwi.json",
                "dwi.bval",
                "dwi.bvec",
            ],
            "task": "",
        },
        "rsfmri": {
            "datatype": "func",
            "suffix": "bold",
            "sidecars": ["rsfmri.json"],
            "task": "task-rest",
        },
        "fieldmap": {
            "datatype": "fmap",
            "suffix": "fmap",
            "sidecars": ["fmap.json"],
            "task": "",
        },
    }
    df = (
        df.assign(
            modality=lambda x: x.series_desc.apply(lambda y: identify_modality(y))
        )
        .dropna(subset=["modality"])
        .drop_duplicates()
    )
    return df.join(df.modality.map(modality_mapping).apply(pd.Series))


def _compute_fieldmaps(df: DataFrame) -> DataFrame:
    """Compute the fieldmaps. Please add details...
    For example, why do we need to compute the dir_num before and after ?
    """
    from clinica.utils.filemanip import get_parent

    df_with_dir_num = df.assign(
        dir_num=lambda x: x.source.apply(
            lambda y: int(get_parent(y).name.split("-")[0])
        )
    )
    df_suf = _merge_fieldmaps(df_with_dir_num, _identify_fieldmaps(df_with_dir_num))
    return df_suf


def _identify_fieldmaps(df: DataFrame) -> DataFrame:
    """Identifies the fieldmaps imaging type: magnitude or phase difference.

    Parameters
    ----------
    df:Dataframe
        Dataframe without the fieldmaps identified.

    Returns
    -------
    Dataframe
        Dataframe with the fieldmaps identified
    """
    column_filter = ["source_id", "source_ses_id", "modality", "dir_num", "suffix"]
    df1 = (
        df[column_filter][df["modality"].str.contains("fieldmap")]
        .groupby(column_filter[:-1])
        .min()
    )
    df2 = (
        df[column_filter[:-1]][df["modality"].str.contains("fieldmap")]
        .groupby(column_filter[:-2])
        .min()
    )
    df1 = df1.join(df2.rename(columns={"dir_num": "run_01_dir_num"}))
    df_alt = df1.reset_index().assign(
        fmap_type_p=lambda x: (x.run_01_dir_num != x.dir_num)
    )
    return df_alt.assign(
        fmap=lambda x: x.fmap_type_p.apply(lambda y: "phasediff" if y else "magnitude")
    )


def _merge_fieldmaps(df: DataFrame, df_fmap: DataFrame) -> DataFrame:
    """Merges the dataframe containing the fieldmaps names

    Parameters
    ----------
    df: Dataframe
        Initial dataframe. Fieldmaps are identified on information extracted here.

    df_fmap: Dataframe
        Dataframe in which fieldmaps are identified.

    Returns
    -------
    Dataframe
        Dataframe containing the correct fieldmap for each fieldmap acquisition.
    """
    column_filter = ["source_id", "source_ses_id", "modality", "dir_num", "fmap"]
    df1 = df.merge(df_fmap[column_filter], how="inner", on=column_filter[:-1])
    df2 = df.merge(df_fmap[column_filter], how="left", on=column_filter[:-1])
    df2 = df2[~df2["modality"].str.contains("fieldmap")]
    df1 = df1.assign(suffix=lambda x: x.fmap)
    return pd.concat([df2, df1], ignore_index=True)


def _compute_runs(df: DataFrame) -> DataFrame:
    """Compute the run numbers."""
    df_with_number_of_parts = _compute_number_of_scan_parts(df)
    df_with_run_numbers = _compute_run_numbers_from_parts(df_with_number_of_parts)
    return df.merge(
        df_with_run_numbers[
            [
                "source_id",
                "source_ses_id",
                "suffix",
                "dir_num",
                "run_num",
                "number_of_parts",
            ]
        ],
        how="left",
        on=["source_id", "source_ses_id", "suffix", "dir_num"],
    )


def _compute_number_of_scan_parts(df: DataFrame) -> DataFrame:
    """Compute the number of parts. Explain me please..."""
    df_alt = _compute_philips_parts(df)
    df_parts = df.merge(
        df_alt[["source_id", "source_ses_id", "suffix", "dir_num", "number_of_parts"]],
        how="left",
        on=["source_id", "source_ses_id", "suffix", "dir_num"],
    )
    df_parts[["number_of_parts"]] = df_parts[["number_of_parts"]].fillna(value="1")
    return df_parts


def _compute_philips_parts(df: DataFrame) -> DataFrame:
    """Compute the parts numbers for philips dwi acquisitions.

    The amount of dwi acquisitions linked together is indicated.
    For example, if a dwi acquisition is split in 9,
    the `number_of_parts` column will have a value of 9 for all of these acquisitions.
    Two columns are added:
        - part_number, which contains the number for each part of a DTI acquisition.
        - number_of_parts, which contains the amount of parts for each DTI acquisition.

    Parameters
    ----------
    df: DataFrame
        Dataframe without runs.

    Returns
    -------
    DataFrame
        Dataframe containing the correct dwi part number for each acquisition. It also contains
        the total amount of dwi parts for each subjects-session.
    """
    df_with_duplicate_flags = _find_duplicate_run(df)
    df_with_part_numbers = _compute_part_numbers(df_with_duplicate_flags)
    df_with_number_of_parts = _compute_number_of_parts(df_with_part_numbers)
    return pd.concat(
        [df_with_part_numbers, df_with_number_of_parts["number_of_parts"]], axis=1
    )


def _find_duplicate_run(df: DataFrame) -> DataFrame:
    """Create a column that contains the information of whether a run is a duplicate or not."""
    column_filter = ["source_id", "source_ses_id", "suffix", "dir_num"]
    df = df[df["suffix"].str.contains("dwi", case=False)]
    df1 = df[column_filter].groupby(column_filter).min()
    df2 = df[column_filter].groupby(column_filter[:-1]).min()
    df1 = df1.join(df2.rename(columns={"dir_num": "part_01_dir_num"}))
    return df1.reset_index().assign(run=lambda x: (x.part_01_dir_num != x.dir_num))


def _compute_part_numbers(df: DataFrame) -> DataFrame:
    """Compute the sequence number of each part."""
    return pd.concat(
        [
            df,
            pd.DataFrame(
                _compute_scan_sequence_numbers(df.run.tolist()), columns=["part_number"]
            ),
        ],
        axis=1,
    )


def _compute_number_of_parts(df: pd.DataFrame) -> DataFrame:
    """Add the number of parts (the max value of the part_number column) to each part."""
    column_filter = ["source_id", "source_ses_id", "suffix", "part_number"]
    df_parts_1 = df[column_filter].groupby(column_filter).max()
    df_parts_2 = df[column_filter].groupby(column_filter[:-1]).max()
    return df_parts_1.join(
        df_parts_2.rename(columns={"part_number": "number_of_parts"})
    ).reset_index()


def _compute_run_numbers_from_parts(df: DataFrame) -> DataFrame:
    """This functions computes the run numbers.

    Parameters
    ----------
    df: DataFrame
        Dataframe without runs.

    Returns
    -------
    DataFrame
        Dataframe containing the correct run for each acquisition.
    """
    column_filter = [
        "source_id",
        "source_ses_id",
        "suffix",
        "number_of_parts",
        "dir_num",
    ]
    df1 = df[column_filter].groupby(column_filter).min()
    df2 = df[column_filter].groupby(column_filter[:-1]).min()
    df1 = df1.join(df2.rename(columns={"dir_num": "run_01_dir_num"}))
    df_alt = df1.reset_index().assign(run=lambda x: (x.run_01_dir_num != x.dir_num))
    df_run = pd.concat(
        [
            df_alt,
            pd.DataFrame(
                _compute_scan_sequence_numbers(df_alt.run.tolist()),
                columns=["run_number"],
            ),
        ],
        axis=1,
    )
    return df_run.assign(run_num=lambda x: x.run_number.apply(lambda y: f"run-{y:02d}"))


def _compute_scan_sequence_numbers(duplicate_flags: Iterable[bool]) -> List[int]:
    """Return the run number from an iterable of booleans indicating
    whether each scan is a duplicate or not.

    Parameters
    ---------
    duplicate_flags : Iterable[bool]
        If the element at index k is True, then the scan k is a duplicate.
        Otherwise, it is the first scan of the sequence.

    Returns
    ------
    ses_numbers : List[int]
        The list of scan sequence numbers.

    Examples
    ---------
    >>> _compute_scan_sequence_numbers([False, True, False, True, False, False, False, True, False, True, True])
    [1, 2, 1, 2, 1, 1, 1, 2, 1, 2, 3]

    Raises
    -----
    ValueError :
        If the input list is empty.
    """
    if len(duplicate_flags) == 0:
        raise ValueError("Provided list is empty.")
    ses_numbers = [1]
    for k in range(1, len(duplicate_flags)):
        ses_numbers.append(1 if not duplicate_flags[k] else ses_numbers[k - 1] + 1)
    return ses_numbers


def _compute_bids_full_path(df: DataFrame) -> DataFrame:
    """Compute the BIDS full path."""
    df_with_bids_filename = df.assign(
        bids_filename=lambda x: x[
            ["participant_id", "session_id", "task", "run_num", "suffix"]
        ].agg("_".join, axis=1),
    )
    df_with_bids_filename = df_with_bids_filename.assign(
        bids_filename=lambda x: x.bids_filename.apply(lambda y: y.replace("__", "_"))
    )
    return df_with_bids_filename.assign(
        bids_full_path=lambda x: x[
            ["participant_id", "session_id", "datatype", "bids_filename"]
        ].agg("/".join, axis=1)
    )


def write_bids(
    to: Path,
    participants: DataFrame,
    sessions: DataFrame,
    scans: DataFrame,
) -> None:
    """This function writes the BIDS

    Parameters
    ----------
    to: Path
        The path where the BIDS should be written

    participants: DataFrame
        DataFrame containing the data for the participants.tsv

    sessions: DataFrame
        DataFrame containing the data for the sessions.tsv

    scans: DataFrame
        DataFrame containing the data for the scans.tsv
    """
    import os

    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
    from clinica.iotools.bids_utils import run_dcm2niix, write_to_tsv
    from clinica.utils.stream import cprint

    cprint("Starting to write the BIDS.", lvl="info")
    fs = LocalFileSystem(auto_mkdir=True)
    # Ensure BIDS hierarchy is written first.

    participants = (
        participants.reset_index()
        .drop(["session_id", "modality", "run_num", "bids_filename", "source"], axis=1)
        .drop_duplicates()
        .set_index("participant_id")
    )
    with fs.transaction:
        with fs.open(to / "dataset_description.json", "w") as dataset_description_file:
            BIDSDatasetDescription(name="GENFI").write(to=dataset_description_file)
        with fs.open(to / "participants.tsv", "w") as participant_file:
            write_to_tsv(participants, participant_file)

    for participant_id, data_frame in sessions.groupby("participant_id"):
        sessions = data_frame.droplevel(
            ["participant_id", "modality", "bids_filename", "run_num"]
        ).drop_duplicates()

        sessions_filepath = to / str(participant_id) / f"{participant_id}_sessions.tsv"
        with fs.open(sessions_filepath, "w") as sessions_file:
            write_to_tsv(sessions, sessions_file)
    scans = scans.reset_index().set_index(["bids_full_path"], verify_integrity=True)
    for bids_full_path, metadata in scans.iterrows():
        bids_full_path = Path(bids_full_path)
        try:
            os.makedirs(to / bids_full_path.parent)
        except OSError:
            pass
        dcm2niix_success = run_dcm2niix(
            Path(metadata["source_path"]).parent,
            to / Path(bids_full_path).parent,
            metadata["bids_filename"],
            True,
        )
        if dcm2niix_success:
            scans_filepath = (
                to
                / str(metadata.participant_id)
                / str(metadata.session_id)
                / f"{metadata.participant_id}_{metadata.session_id}_scan.tsv"
            )
            row_to_write = _serialize_row(
                metadata.drop(["participant_id", "session_id"]),
                write_column_names=not scans_filepath.exists(),
            )
            with open(scans_filepath, "a") as scans_file:
                scans_file.write(f"{row_to_write}\n")
            if (
                "dwi" in metadata["bids_filename"]
                and "Philips" in metadata.manufacturer
            ):
                _merge_philips_diffusion(
                    to / bids_full_path.with_suffix(".json"),
                    metadata.number_of_parts,
                    metadata.run_num,
                )
    _correct_fieldmaps_name(to)
    _delete_real_and_imaginary_files(to)


def _serialize_row(row: pd.Series, write_column_names: bool) -> str:
    row_dict = row.to_dict()
    to_write = (
        [row_dict.keys(), row_dict.values()]
        if write_column_names
        else [row_dict.values()]
    )
    return "\n".join([_serialize_list(list(_)) for _ in to_write])


def _serialize_list(data: list, sep="\t") -> str:
    return sep.join([str(value) for value in data])


def _correct_fieldmaps_name(to: Path) -> None:
    """This function scans the BIDS after it has been written to correct the nameds of the fieldmap files.

    Parameters
    ----------
    to: Path
        The path to the BIDS

    Examples
    --------
    bids
    ├── README
    ├── dataset_description.json
    ├── participants.tsv
    └── sub-GRN001
        ├── ses-M000
        │   ├── fmap
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude_e1.json
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude_e1.nii.gz
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude_e2.json
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude_e2.nii.gz
        │   │   ├── sub-GRN001_ses-M000_run-01_phasediff_e2_ph.json
        │   │   └── sub-GRN001_ses-M000_run-01_phasediff_e2_ph.nii.gz
        │   └── func
        │       ├── sub-GRN001_ses-M000_run-01_bold.json
        │       └── sub-GRN001_ses-M000_run-01_bold.nii.gz
        └── sub-GRN001_sessions.tsv
    >>> _correct_fieldmaps_name("path/to/bids)
    bids
    ├── README
    ├── dataset_description.json
    ├── participants.tsv
    └── sub-GRN001
        ├── ses-M000
        │   ├── fmap
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude1.json
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude1.nii.gz
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude2.json
        │   │   ├── sub-GRN001_ses-M000_run-01_magnitude2.nii.gz
        │   │   ├── sub-GRN001_ses-M000_run-01_phasediff.json
        │   │   └── sub-GRN001_ses-M000_run-01_phasediff.nii.gz
        │   └── func
        │       ├── sub-GRN001_ses-M000_run-01_bold.json
        │       └── sub-GRN001_ses-M000_run-01_bold.nii.gz
        └── sub-GRN001_sessions.tsv
    """
    import os
    import re

    for z in to.rglob("*magnitude_e*"):
        os.rename(z, z.parent / re.sub(r"magnitude_e", "magnitude", z.name))

    for z in to.rglob("*phasediff_e*_ph*"):
        os.rename(z, z.parent / re.sub(r"phasediff_e[1-9]_ph", "phasediff", z.name))


def _merge_philips_diffusion(
    json_file: Path,
    number_of_parts: float,
    run_num: str,
) -> None:
    """Add the dwi number in the provided JSON file for each run of Philips images.
    The 'MultipartID' field of the input JSON file is set to 'dwi_1' or 'dwi_2' depending
    on the run number.
    """
    import json

    from clinica.utils.stream import cprint

    multipart_id = _get_multipart_id(
        PhilipsNumberOfParts.from_int(int(number_of_parts)), run_num
    )
    try:
        with open(json_file, "r+") as f:
            if multipart_id is not None:
                data = json.load(f)
                data["MultipartID"] = multipart_id
                json.dump(data, f, indent=4)
    except FileNotFoundError:
        cprint(msg=f"the file {json_file} does not exist.", lvl="warning")


class PhilipsNumberOfParts(Enum):
    """DWI scans obtained with a Philips scanner might have
    been divided in either 5 or 9 parts. This distinction is important
    because we will link these different parts together.
    If the number of parts is not 5 or 9, nothing will be done.
    """

    FIVE = 5
    NINE = 9
    OTHER = None

    @classmethod
    def from_int(cls, nb_parts: int):
        import warnings

        for member in cls:
            if member.value == nb_parts:
                return member
        warnings.warn(
            f"Unexpected number of splits {nb_parts}. "
            f"Should be one of {[c.value for c in cls]}."
        )
        return cls.OTHER


def _get_multipart_id(nb_parts: PhilipsNumberOfParts, run_num: str) -> Optional[str]:
    """Return the MultiPartID for the provided run number depending on the number of parts."""
    if nb_parts == PhilipsNumberOfParts.NINE:
        if run_num in (f"run-0{k}" for k in range(1, 5)):
            return "dwi_2"
        if run_num in (f"run-0{k}" for k in range(5, 10)):
            return "dwi_1"
        raise ValueError(f"{run_num} is outside of the scope.")
    if nb_parts == PhilipsNumberOfParts.FIVE:
        if run_num in (f"run-0{k}" for k in range(1, 6)):
            return "dwi_1"
        raise ValueError(f"{run_num} is outside of the scope.")
    if nb_parts == PhilipsNumberOfParts.OTHER:
        return None


def _delete_real_and_imaginary_files(bids_folder: Path) -> None:
    """Delete all files with a `_real` or `_imaginary` suffix from the BIDS folder.
    These files are generated by dcm2niix when it is not able to correctly convert the full image.
    These files break the BIDS specification, and are therefore removed.

    Parameters
    ----------
    bids_folder : PathLike
        Path to the BIDS folder.
    """
    from clinica.utils.stream import cprint

    for file_type in ("real", "imaginary"):
        for f in bids_folder.rglob(f"*_{file_type}.*"):
            f.unlink()
            cprint(f"file {f} was deleted as it is not BIDS compliant.")



"""This module contains utilities to handle atlases in Clinica.

An atlas is currently defined by its name and a set of labels in a template space.

This current implementation has some drawbacks:
- Atlas is misleading: it is only a set of labels in a template space
- This implementation can not handle case where parcellation is in several template space
(e.g. `MNI152NLin2009cAsym` and `MNI152NLin6Asym`, see TemplateFlow for example)

Either a refactoring of this module or the use of an external API
(e.g. TemplateFlow - https://www.templateflow.org/) needs to be considered.
"""

import abc
from enum import Enum
from pathlib import Path
from typing import Union

import nibabel as nib
import numpy as np
from nibabel import Nifti1Header


class T1AndPetVolumeAtlasName(str, Enum):
    """Possible names for T1 / PET atlases."""

    AAL2 = "AAL2"
    AICHA = "AICHA"
    HAMMERS = "Hammers"
    LPBA40 = "LPBA40"
    NEUROMORPHOMETRICS = "Neuromorphometrics"


class AtlasName(str, Enum):
    """Possible names for atlases."""

    AAL2 = "AAL2"
    AICHA = "AICHA"
    HAMMERS = "Hammers"
    LPBA40 = "LPBA40"
    NEUROMORPHOMETRICS = "Neuromorphometrics"
    JHUDTI81 = "JHUDTI81"
    JHUTract0 = "JHUTracts0"
    JHUTract25 = "JHUTracts25"
    JHUTracts50 = "JHUTracts50"


def _get_resolution_along_axis(label_image_header: Nifti1Header, axis: int) -> str:
    voxels_labels = label_image_header.get_zooms()
    if not 0 <= axis < len(voxels_labels):
        raise ValueError(
            f"The label image has dimension {len(voxels_labels)} and "
            f"axis {axis} is therefore not valid. Please use a value "
            f"between 0 and {len(voxels_labels) - 1}."
        )
    if int(voxels_labels[axis]) == voxels_labels[axis]:
        return str(int(voxels_labels[axis]))
    return str(voxels_labels[axis])


class BaseAtlas:
    """Base class for Atlas handling."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name: str, roi_filename: str):
        self.name = name
        self.roi_filename = roi_filename
        self.atlas_dir = None
        self.atlas_filename = None

    @property
    @abc.abstractmethod
    def expected_checksum(self) -> str:
        raise NotImplementedError

    @property
    def atlas_folder(self) -> Path:
        return Path(__file__).parent.parent / "resources" / "atlases"

    @property
    def tsv_roi(self) -> Path:
        """Path to the parcellation TSV file.

        The TSV file must contain the `roi_value` and `roi_name` columns:

        roi_value   roi_name
        0   Background
        2001    Precentral_L
        [...]   [...]
        9170    Vermis_10
        """
        return self.atlas_folder / self.roi_filename

    @property
    def spatial_resolution(self) -> str:
        """Return the spatial resolution of the atlas (in format "XxXxX" e.g. 1.5x1.5x1.5)."""
        img_labels = nib.load(self.labels)
        return "x".join(
            _get_resolution_along_axis(img_labels.header, axis=axis)
            for axis in range(3)
        )

    @property
    def labels(self) -> Path:
        """Path to the parcellation in NIfTI format.

        Raises
        ------
        IOError :
            If the checksum of the parcellation found is different from
            the expected checksum.
        """
        from .inputs import compute_sha256_hash

        atlas_labels = self.atlas_dir / self.atlas_filename
        if (checksum := compute_sha256_hash(atlas_labels)) != self.expected_checksum:
            raise IOError(
                f"{atlas_labels} has an SHA256 checksum ({checksum}) "
                f"differing from expected ({self.expected_checksum}), "
                f"file may be corrupted and changed with newer version of FSL."
            )
        return atlas_labels

    def get_index(self) -> np.ndarray:
        img_labels = nib.load(self.labels)
        img_labels = img_labels.get_fdata(dtype="float32")
        labels = list(set(img_labels.ravel()))
        index_vector = np.zeros(len(labels))
        for index, n in enumerate(labels):
            index_vector[index] = index
        return index_vector


class FSLAtlas(BaseAtlas):
    """FSL atlases look for the labels in the FSL folder (requires FSL)."""

    def __init__(self, name: str, roi_filename: str, atlas_filename: str):
        from .check_dependency import get_fsl_home

        super().__init__(name, roi_filename)
        self.atlas_filename = atlas_filename
        self.atlas_dir = get_fsl_home() / "data" / "atlases" / "JHU"


class JHUDTI811mm(FSLAtlas):
    """JHUDTI811mm atlas.

    This atlas contains 48 white matter tract labels that were created by manually
    segmenting a standard-space average of diffusion MRI tensor maps from 81 subjects.

    References
    ----------
    https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/Atlases
    https://www.sciencedirect.com/science/article/abs/pii/S105381190700688X?via%3Dihub
    """

    def __init__(self):
        super().__init__(
            name="JHUDTI81",
            roi_filename="atlas-JHUDTI81_dseg.tsv",
            atlas_filename="JHU-ICBM-labels-1mm.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        import nipype.interfaces.fsl as fsl

        if ["5", "0", "5"] <= fsl.Info.version().split(".") < ["6", "0", "5"]:
            return "fac584ec75ff2a8631710d3345df96733ed87d9bde3387f5b462f8d22914ed69"
        return "3c3f5d2f1250a3df60982acff35a75b99fd549a05d5f8124a63f78221aa0ec16"


class JHUTracts01mm(FSLAtlas):
    """JHUTracts01mm atlas.

    This atlas contains 20 white matter tract labels that were identified probabilistically
    by averaging the results of deterministic tractography run on 28 subjects.
    Threshold used is 0%.

    References
    ----------
    https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/Atlases
    https://shop.elsevier.com/books/mri-atlas-of-human-white-matter/mori/978-0-444-51741-8
    """

    def __init__(self):
        super().__init__(
            name="JHUTracts0",
            roi_filename="atlas-JHUTract_dseg.tsv",
            atlas_filename="JHU-ICBM-tracts-maxprob-thr0-1mm.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "eb1de9413a46b02d2b5c7b77852097c6f42c8a5d55a5dbdef949c2e63b95354e"


class JHUTracts251mm(FSLAtlas):
    """JHUTracts251mm atlas.

    This atlas contains 20 white matter tract labels that were identified probabilistically
    by averaging the results of deterministic tractography run on 28 subjects.
    Threshold used is 25%.

    References
    ----------
    https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/Atlases
    https://shop.elsevier.com/books/mri-atlas-of-human-white-matter/mori/978-0-444-51741-8
    """

    def __init__(self):
        super().__init__(
            name="JHUTracts25",
            roi_filename="atlas-JHUTract_dseg.tsv",
            atlas_filename="JHU-ICBM-tracts-maxprob-thr25-1mm.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "7cd85fa2be1918fc83173e9bc0746031fd4c08d70d6c81b7b9224b5d3da6d8a6"


class JHUTracts501mm(FSLAtlas):
    """JHUTracts501mm atlas.

    This atlas contains 20 white matter tract labels that were identified probabilistically
    by averaging the results of deterministic tractography run on 28 subjects.
    Threshold used is 50%.

    References
    ----------
    https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/Atlases
    https://shop.elsevier.com/books/mri-atlas-of-human-white-matter/mori/978-0-444-51741-8
    """

    def __init__(self):
        super().__init__(
            name="JHUTracts50",
            roi_filename="atlas-JHUTract_dseg.tsv",
            atlas_filename="JHU-ICBM-tracts-maxprob-thr50-1mm.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "20ff0216d770686838de26393c0bdac38c8104760631a1a2b5f518bc0bbb470a"


class LocalAtlas(BaseAtlas):
    """Local atlases will look for labels in the local 'resources' folder.

    More precisely, the labels and TSV files associated with the atlas
    are located in the folder `<clinica>/resources/atlases/`.
    """

    def __init__(self, name: str, roi_filename: str, atlas_filename: str):
        super().__init__(name, roi_filename)
        self.atlas_filename = atlas_filename
        self.atlas_dir = Path(__file__).parent.parent / "resources" / "atlases"


class AAL2(LocalAtlas):
    """AAL2 atlas.

    Anatomical atlas based on a single subject. It is the updated version of AAL, which is
    probably the most widely used cortical parcellation map in the neuroimaging literature.
    It was built using manual tracing on the spatially normalized single-subject high-resolution
    T1 volume in MNI space. It is composed of 120 regions covering the whole cortex as well as
    the main subcortical structures.

    References
    ----------
    https://www.gin.cnrs.fr/en/tools/aal/
    https://www.sciencedirect.com/science/article/abs/pii/S1053811901909784?via%3Dihub
    """

    def __init__(self):
        super().__init__(
            name="AAL2",
            roi_filename="atlas-AAL2_dseg.tsv",
            atlas_filename="atlas-AAL2_dseg.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "f6bc698f778a4b383abd3ce355bfd4505c4aa14708e4a7848f8ee928c2b56b37"


class AICHA(LocalAtlas):
    """AICHA atlas.

    Functional atlas based on multiple subjects. It was built using parcellation of group-level
    functional connectivity profiles computed from resting-state fMRI data of 281 healthy subjects.
    It is composed of 384 regions covering the whole cortex as well as the main subcortical structures.

    References
    ----------
    https://www.gin.cnrs.fr/en/tools/aicha/
    https://www.sciencedirect.com/science/article/abs/pii/S0165027015002678?via%3Dihub
    """

    def __init__(self):
        super().__init__(
            name="AICHA",
            roi_filename="atlas-AICHA_dseg.tsv",
            atlas_filename="atlas-AICHA_dseg.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "cab554d5f546720e60f61f536f82c3d355b31fadb5a4d3ce6a050a606d7ef761"


class RemoteAtlas(BaseAtlas):
    """Remote atlases will download the labels from the aramislab server."""

    def __init__(self, name: str, roi_filename: str, atlas_filename: str):
        super().__init__(name, roi_filename)
        self.atlas_filename = atlas_filename

    @property
    def labels(self) -> Path:
        from clinica.utils.inputs import RemoteFileStructure, get_file_from_server

        return get_file_from_server(
            RemoteFileStructure(
                filename=self.atlas_filename,
                url="https://aramislab.paris.inria.fr/files/software/cat12/CAT12-Atlases/",
                checksum=self.expected_checksum,
            )
        )


class Hammers(RemoteAtlas):
    """Hammers atlas.

    Anatomical atlas based on multiple subjects. It was built using manual tracing on anatomical
    MRI from 30 healthy subjects. The individual subjects parcellations were then registered to MNI
    space to generate a probabilistic atlas as well as a maximum probability map. The latter was
    used in the present work. It is composed of 69 regions covering the whole cortex as well
    as he main subcortical structures.

    References
    ----------
    https://neuro-jena.github.io/cat//index.html#DOWNLOAD
    https://onlinelibrary.wiley.com/doi/epdf/10.1002/hbm.10123
    """

    def __init__(self):
        super().__init__(
            name="Hammers",
            roi_filename="atlas-Hammers_dseg.tsv",
            atlas_filename="atlas-Hammers_dseg.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "c034a7bce2dcab390a0b72f4e7d04769eb3fe5b990d0e18d89b0ce73339a5376"


class LPBA40(RemoteAtlas):
    """LPBA40 atlas.

    Anatomical atlas based on multiple subjects. It was built using manual tracing on anatomical
    MRI from 40 healthy subjects. The individual subjects parcellations were then registered to MNI
    space to generate a maximum probability map. It is composed of 56 regions covering the whole
    cortex as well as the main subcortical structures.

    References
    ----------
    https://neuro-jena.github.io/cat//index.html#DOWNLOAD
    https://www.sciencedirect.com/science/article/abs/pii/S1053811907008099?via%3Dihub
    """

    def __init__(self):
        super().__init__(
            name="LPBA40",
            roi_filename="atlas-LPBA40_dseg.tsv",
            atlas_filename="atlas-LPBA40_dseg.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "20826b572bbbdbcdbf28bbd3801dc0c2fed28d1e54bc4fd5027e64ccc6d50374"


class Neuromorphometrics(RemoteAtlas):
    """Neuromorphometrics atlas.

    Anatomical atlas based on multiple subjects. It was built using manual tracing on anatomical
    MRI from 30 healthy subjects. The individual subjects parcellations were then registered to
    MNI space to generate a maximum probability map. It is composed of 140 regions covering the
    whole cortex as well as the main subcortical structures. Data were made available for the
    “MICCAI 2012 Grand Challenge and Workshop on Multi-Atlas Labeling”.

    References
    ----------
    https://neuro-jena.github.io/cat//index.html#DOWNLOAD
    http://masiweb.vuse.vanderbilt.edu/workshop2012/index.php/Challenge_Details
    """

    def __init__(self):
        super().__init__(
            name="Neuromorphometrics",
            roi_filename="atlas-Neuromorphometrics_dseg.tsv",
            atlas_filename="atlas-Neuromorphometrics_dseg.nii.gz",
        )

    @property
    def expected_checksum(self) -> str:
        return "19a50136cd2f8a14357a19ad8a1dc4a2ecb6beb3fc16cb5441f4f2ebaf64a9a5"


def atlas_factory(atlas_name: Union[str, AtlasName, BaseAtlas]) -> BaseAtlas:
    """Factory method for atlases.

    Parameters
    ----------
    atlas_name : str or AtlasName or atlas instance
        If an atlas instance, the instance is returned.
        If a string, the corresponding atlas will be returned.

    Returns
    -------
    BaseAtlas :
        The atlas class corresponding to the provided name.
    """
    if isinstance(atlas_name, BaseAtlas):
        return atlas_name
    if isinstance(atlas_name, str):
        atlas_name = AtlasName(atlas_name)
    if atlas_name == AtlasName.AAL2:
        return AAL2()
    if atlas_name == AtlasName.AICHA:
        return AICHA()
    if atlas_name == AtlasName.HAMMERS:
        return Hammers()
    if atlas_name == AtlasName.LPBA40:
        return LPBA40()
    if atlas_name == AtlasName.NEUROMORPHOMETRICS:
        return Neuromorphometrics()
    if atlas_name == AtlasName.JHUDTI81:
        return JHUDTI811mm()
    if atlas_name == AtlasName.JHUTract0:
        return JHUTracts01mm()
    if atlas_name == AtlasName.JHUTract25:
        return JHUTracts251mm()
    if atlas_name == AtlasName.JHUTracts50:
        return JHUTracts501mm()



from string import Template
from typing import Optional

import pandas as pd
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect

__all__ = ["build_model", "print_clusters", "check_column_in_df", "is_categorical"]


MISSING_TERM_ERROR_MSG = Template(
    "Term ${term} from the design matrix is not in the columns of the "
    "provided TSV file. Please make sure that there is no typo."
)


def print_clusters(model: SLM, threshold: float) -> None:
    """This function prints the results related to total number
    of clusters, as well as the significative clusters.

    Parameters
    ----------
    model : brainstat.stats.SLM
        Fitted SLM model.

    threshold : float
        Cluster defining threshold.
    """
    from clinica.utils.stream import cprint

    cprint("#" * 40)
    cprint("After correction (Cluster-wise Correction for Multiple Comparisons): ")
    df = model.P["clus"][0]
    cprint(df)
    cprint(f"Clusters found: {len(df)}")
    cprint(
        f"Significative clusters (after correction): {len(df[df['P'] <= threshold])}"
    )


def build_model(design_matrix: str, df: pd.DataFrame) -> FixedEffect:
    """Build a brainstat model from the design matrix in string format.

    This function assumes that the design matrix is formatted
    in the following way:

        1 + factor_1 + factor_2 + ...

    Or:

        factor_1 + factor_2 + ...

    in the latter case the intercept will be added automatically.

    Parameters
    ----------
    design_matrix : str
        Design matrix specified as a string.

    df : pd.DataFrame
        Subjects DataFrame.

    Returns
    -------
    model : FixedEffect
        BrainStats model.
    """
    from functools import reduce

    if len(design_matrix) == 0:
        raise ValueError("Design matrix cannot be empty.")
    if "+" in design_matrix:
        terms = [_.strip() for _ in design_matrix.split("+")]
    else:
        terms = [design_matrix.strip()]
    model = []
    for term in terms:
        # Intercept is automatically included in brainstat
        if term == "1":
            continue
        # Handles the interaction effects
        if "*" in term:
            sub_terms = [_.strip() for _ in term.split("*")]
            model_term = reduce(
                lambda x, y: x * y, [_build_model_term(_, df) for _ in sub_terms]
            )
        else:
            model_term = _build_model_term(term, df)
        model.append(model_term)
    return reduce(lambda x, y: x + y, model)


def _build_model_term(
    term: str,
    df: pd.DataFrame,
    add_intercept: Optional[bool] = True,
) -> FixedEffect:
    """Builds a BrainStats model term from the subjects
    DataFrame and a column name.

    Parameters
    ----------
    term : str
        The name of the column of the DataFrame to be used.

    df : pd.DataFrame
        The subjects DataFrame.

    add_intercept : bool
        If `True`, adds an intercept term.

    Returns
    -------
    FixedEffect :
        BrainStats model term.
    """
    return FixedEffect(df[term], add_intercept=add_intercept)


def check_column_in_df(df: pd.DataFrame, column: str) -> None:
    """Checks if the provided column name is in the provided DataFrame.
    Raises a ValueError if not.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to analyze.

    column : str
        Name of the column to check.
    """
    if column not in df.columns:
        raise ValueError(MISSING_TERM_ERROR_MSG.safe_substitute(term=column))


def is_categorical(df: pd.DataFrame, column: str) -> bool:
    """Returns `True` if the column is categorical and `False` otherwise.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to analyze.

    column : str
        The name of the column to check.

    Returns
    -------
    bool :
        `True` if the column contains categorical values, `False` otherwise.
    """
    return not df[column].dtype.name.startswith("float")



import warnings
from typing import List, Optional

import pandas as pd

from ._base import GLM
from ._contrast import GroupContrast, GroupContrastWithInteraction


class GroupGLM(GLM):
    """Class implementing group GLM models.

    Attributes
    ----------
    See documentation for `GLM` class.

    group_label : str, optional
        The Label to use for group GLM models. Default="group".
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        group_label: Optional[str] = "group",
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        super().__init__(
            design,
            df,
            feature_label,
            contrast,
            fwhm,
            threshold_uncorrected_pvalue,
            threshold_corrected_pvalue,
            cluster_threshold,
        )
        self.with_interaction = False
        self.group_label = group_label

    def _build_contrasts(
        self, contrast: str, subjects_df: pd.DataFrame
    ) -> List[GroupContrast]:
        return [
            GroupContrast.from_string(contrast, subjects_df, sign)
            for sign in ("positive", "negative")
        ]

    def _get_output_filename(self, contrast: GroupContrast) -> str:
        """Build the filename root part from class attributes and provided contrast.

        Parameters
        ----------
        contrast : GroupContrast
            The contrast to use for building the filename.
        """
        return f"group-{self.group_label}_{contrast.name}_measure-{self.feature_label}_fwhm-{self.fwhm}"


class GroupGLMWithInteraction(GroupGLM):
    """This class implements a GLM model for group comparison with
    interaction effects.

    Attributes
    ----------
    See attributes of parent class `GroupGLM`.
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        group_label: Optional[str] = "group",
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        super().__init__(
            design,
            df,
            feature_label,
            contrast,
            group_label,
            fwhm,
            threshold_uncorrected_pvalue,
            threshold_corrected_pvalue,
            cluster_threshold,
        )
        self.with_interaction: bool = True
        warnings.warn(
            "You included interaction as covariate in your model, "
            "please carefully check the format of your tsv files."
        )

    def _build_contrasts(
        self, contrast: str, subjects_df: pd.DataFrame
    ) -> List[GroupContrastWithInteraction]:
        return [GroupContrastWithInteraction.from_string(contrast, subjects_df)]

    def _get_output_filename(self, contrast: GroupContrastWithInteraction) -> str:
        """Build the filename from class attributes and provided contrast.

        Parameters
        ----------
        contrast : GroupContrastWithInteraction
            The contrast to use for building the filename.
        """
        return (
            f"interaction-{contrast.name}_measure-{self.feature_label}_fwhm-{self.fwhm}"
        )
